# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from enum import Enum
from typing import Any, Dict, List, Optional

import nncf
from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.common.utils.api_marker import api
from nncf.config.schemata.defaults import QUANTIZATION_BITS
from nncf.config.schemata.defaults import QUANTIZATION_PER_CHANNEL
from nncf.parameters import StrEnum
from nncf.parameters import TargetDevice


@api()
class QuantizationScheme:
    """
    Basic enumeration for quantization scheme specification.

    :param SYMMETRIC:
    :param ASYMMETRIC:
    """

    SYMMETRIC = "symmetric"
    ASYMMETRIC = "asymmetric"


class QuantizerConfig:
    """
    A generic, framework-agnostic information on a configuration of a quantizer for abstract reasoning
    and determination of a quantizer setup scheme for a given model.
    """

    def __init__(
        self,
        num_bits: int = QUANTIZATION_BITS,
        mode: QuantizationScheme = QuantizationScheme.SYMMETRIC,
        signedness_to_force: Optional[bool] = None,
        per_channel: bool = QUANTIZATION_PER_CHANNEL,
    ):
        """
        :param num_bits: Bitwidth of the quantization.
        :param mode: The mode of quantization (symmetric or asymmetric).
        :param signedness_to_force: True if the quantizer *must* be signed, False if *must* be unsigned,
            None if the signed/unsigned attribute should be determined based on the incoming activation
            statistics during range initialization.
        :param per_channel: True for per-channel quantization, False for per-tensor.
        """
        self.num_bits = num_bits
        self.mode = mode
        self.signedness_to_force = signedness_to_force
        self.per_channel = per_channel

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __str__(self):
        return "B:{bits} M:{mode} SGN:{signedness} PC:{per_channel}".format(
            bits=self.num_bits,
            mode="S" if self.mode == QuantizationScheme.SYMMETRIC else "A",
            signedness="ANY" if self.signedness_to_force is None else ("S" if self.signedness_to_force else "U"),
            per_channel="Y" if self.per_channel else "N",
        )

    def __hash__(self):
        return hash(str(self))

    def is_valid_requantization_for(self, other: "QuantizerConfig") -> bool:
        """
        Quantizer config A is a valid requantization for quantizer config B if A is more strict -
        specifically, it might be reasonable to put quantizer A after quantizer B in tensor data control flow, so that
        the requantization will further constrain the input tensor data w.r.t. values it can take, but
        putting quantizer A after quantizer B would be unreasonable.

        :param other: The "primary" QuantizerConfig, i.e. the one that defines an already present quantization.
        :return: True if the current config is a valid requantization for `other`, False otherwise.
        """
        fail_conditions = [
            self.num_bits > other.num_bits,
            self.mode is QuantizationScheme.ASYMMETRIC and other.mode is QuantizationScheme.SYMMETRIC,
            self.signedness_to_force is None and other.signedness_to_force is not None,
            self.signedness_to_force is True and other.signedness_to_force is False,
        ]
        if any(fail_conditions):
            return False
        return True

    def compatible_with_a_unified_scale_linked_qconfig(self, linked_qconfig: "QuantizerConfig"):
        """
        For two configs to be compatible in a unified scale scenario, all of their fundamental parameters
        must be aligned.

        :param linked_qconfig: A QuantizerConfig that is compared against the current config.
        :return: A boolean value specifying whether `linked_qconfig` is compatible with the current config in terms
            of scale unification.
        """
        return (
            self.num_bits == linked_qconfig.num_bits
            and self.mode == linked_qconfig.mode
            and self.signedness_to_force == linked_qconfig.signedness_to_force
            and self.per_channel == linked_qconfig.per_channel
        )

    def is_a_bitwidth_variant(self, other_qconfig: "QuantizerConfig") -> bool:
        """
        :param other_qconfig: A QuantizerConfig to be compared against the current config.
        :return: A boolean value specifying whether `other_config` is identical to the current config
            in everything except the bitwidth.
        """
        return (
            self.per_channel == other_qconfig.per_channel
            and self.signedness_to_force == other_qconfig.signedness_to_force
            and self.mode == other_qconfig.mode
        )

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            "num_bits": self.num_bits,
            "mode": self.mode,
            "signedness_to_force": self.signedness_to_force,
            "per_channel": self.per_channel,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "QuantizerConfig":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return cls(**state)


class QuantizerSpec:
    """
    A specific (potentially framework-aware) parameter struct required to initialize a
    given object that performs quantization of an input tensor.
    """

    def __init__(
        self, num_bits: int, mode: QuantizationScheme, signedness_to_force: bool, narrow_range: bool, half_range: bool
    ):
        """
        :param num_bits: Bitwidth of the quantization.
        :param mode: The mode of quantization (symmetric or asymmetric).
        :param signedness_to_force: True if the quantizer *must* be signed, False if *must* be unsigned,
            None if the signed/unsigned attribute should be determined based on the incoming activation
            statistics during range initialization.
        :param narrow_range: True if the range of quantized values should be narrowed as compared to the
            naive case, False if all 2^`num_bits` quantizations should be used.
        :param half_range: If ``True`` effectively only a half of an quantizer range are used.
            False - the full range are used.
        """
        self.num_bits = num_bits
        self.mode = mode
        self.signedness_to_force = signedness_to_force
        self.narrow_range = narrow_range
        self.half_range = half_range

    def __eq__(self, other: "QuantizerSpec"):
        return self.__dict__ == other.__dict__

    @classmethod
    def from_config(cls, qconfig: QuantizerConfig, narrow_range: bool, half_range: bool) -> "QuantizerSpec":
        return cls(qconfig.num_bits, qconfig.mode, qconfig.signedness_to_force, narrow_range, half_range)


class QuantizationConstraints:
    REF_QCONF_OBJ = QuantizerConfig()

    def __init__(self, **kwargs):
        """
        Use attribute names of QuantizerConfig as arguments
        to set up constraints.
        E.g. QuantizationConstraint(bits=8, per_channel=True) will set up
        a constraint that corresponds to all 8-bit per-channel quantizers, either
        symmetric or asymmetric, either signed or unsigned.
        """
        for attr_name in kwargs:
            if not hasattr(QuantizationConstraints.REF_QCONF_OBJ, attr_name):
                raise nncf.ValidationError(
                    "Invalid constraint - QuantizerConfig has no attribute '{}'".format(attr_name)
                )
        self.qconf_attr_vs_constraint_dict = kwargs

    def apply_constraints_to(self, qconfig: QuantizerConfig) -> QuantizerConfig:
        for attr_name, constraint in self.qconf_attr_vs_constraint_dict.items():
            if constraint is not None:
                setattr(qconfig, attr_name, constraint)
        return qconfig

    def is_config_compatible(self, qconfig: QuantizerConfig) -> bool:
        for attr_name, constraint in self.qconf_attr_vs_constraint_dict.items():
            if constraint is not None:
                qconf_attr_value = getattr(qconfig, attr_name)
                if qconf_attr_value != constraint:
                    return False
        return True

    def get_updated_constraints(self, overriding_constraints: "QuantizationConstraints") -> "QuantizationConstraints":
        new_dict = deepcopy(self.qconf_attr_vs_constraint_dict)
        new_dict.update(overriding_constraints.qconf_attr_vs_constraint_dict)
        return QuantizationConstraints(**new_dict)

    @classmethod
    def from_config_dict(cls, config_dict: Dict) -> "QuantizationConstraints":
        return cls(
            num_bits=config_dict.get("bits"),
            mode=config_dict.get("mode"),
            per_channel=config_dict.get("per_channel"),
            signedness_to_force=config_dict.get("signed"),
        )

    def constrain_qconfig_list(
        self, node_name: NNCFNodeName, target_device: TargetDevice, quantizer_config_list: List[QuantizerConfig]
    ) -> List[QuantizerConfig]:
        assert quantizer_config_list is not None

        constrained_quantizer_config_list = list(filter(self.is_config_compatible, quantizer_config_list))

        # TODO: Make the logic more flexible when the flag "warning as error" is implemented.
        # It means that the qconfig from overrides must be selected as final config
        # even if it is not valid in hw-config.
        if not constrained_quantizer_config_list:
            err_msg = f"Quantization parameter constraints specified in NNCF config are incompatible \
            with HW capabilities as specified in HW config type '{target_device}'. \
            First conflicting quantizer location: {node_name}"
            raise ValueError(err_msg)

        return constrained_quantizer_config_list


class QuantizerGroup(Enum):
    ACTIVATIONS = "activations"
    WEIGHTS = "weights"


class QuantizableWeightedLayerNode:
    def __init__(self, node: NNCFNode, qconfig_list: List[QuantizerConfig]):
        self.node = node
        self.qconfig_list = qconfig_list


class QuantizerId:
    """
    Unique identifier of a quantizer. It's used to store and search all quantizers in a single
    structure.
    """

    def get_base(self):
        raise NotImplementedError

    def get_suffix(self) -> str:
        raise NotImplementedError

    def __str__(self):
        return str(self.get_base()) + self.get_suffix()

    def __hash__(self):
        return hash((self.get_base(), self.get_suffix()))

    def __eq__(self, other: "QuantizerId"):
        return (self.get_base() == other.get_base()) and (self.get_suffix() == other.get_suffix())


class WeightQuantizerId(QuantizerId):
    """Unique identifier of a quantizer for weights."""

    def __init__(self, target_node_name: NNCFNodeName):
        self.target_node_name = target_node_name

    def get_base(self) -> str:
        return self.target_node_name

    def get_suffix(self) -> str:
        return "|WEIGHT"


class NonWeightQuantizerId(QuantizerId):
    """
    Unique identifier of a quantizer, which corresponds to non-weight operations, such as
    ordinary activation, function and input
    """

    def __init__(self, target_node_name: NNCFNodeName, input_port_id=None):
        self.target_node_name = target_node_name
        self.input_port_id = input_port_id

    def get_base(self) -> str:
        return self.target_node_name

    def get_suffix(self) -> str:
        return "|OUTPUT" if self.input_port_id is None else "|INPUT{}".format(self.input_port_id)


class UnifiedScaleType(Enum):
    """
    UNIFY_ONLY_PER_TENSOR - only results in scale unification if per-tensor quantization is ultimately applied.
    This is the target scenario for concat unified scales since the channel count between the concatenated tensors
    may be mismatching and, more importantly, the concatenation might occur on exactly the channel dimension which
    means that the concatenated tensor must reuse all quantization scales of the input per-channel
    quantized tensors.
    UNIFY_ALWAYS - results in scale unification for both per-channel and per-tensor quantization. This is the
    target scenario for eltwise unified scales, as it is assumed that the eltwise ops have matching input
    tensor shapes and therefore the quantization channel count is the same.
    """

    UNIFY_ONLY_PER_TENSOR = 0
    UNIFY_ALWAYS = 1


@api(canonical_alias="nncf.QuantizationPreset")
class QuantizationPreset(StrEnum):
    """
    An enum with values corresponding to the available quantization presets.
    """

    PERFORMANCE = "performance"
    MIXED = "mixed"

    def get_params_configured_by_preset(self, quant_group: QuantizerGroup) -> Dict:
        if quant_group == QuantizerGroup.ACTIVATIONS and self == QuantizationPreset.MIXED:
            return {"mode": QuantizationScheme.ASYMMETRIC}
        return {"mode": QuantizationScheme.SYMMETRIC}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List

from nncf.common.graph.definitions import NNCFGraphNodeType
from nncf.common.pruning.operations import BatchNormPruningOp
from nncf.common.pruning.operations import ConcatPruningOp
from nncf.common.pruning.operations import ConvolutionPruningOp
from nncf.common.pruning.operations import ElementwisePruningOp
from nncf.common.pruning.operations import FlattenPruningOp
from nncf.common.pruning.operations import IdentityMaskForwardPruningOp
from nncf.common.pruning.operations import InputPruningOp
from nncf.common.pruning.operations import LinearPruningOp
from nncf.common.pruning.operations import OutputPruningOp
from nncf.common.pruning.operations import ReshapePruningOp
from nncf.common.pruning.operations import SplitPruningOp
from nncf.common.pruning.operations import StopMaskForwardPruningOp
from nncf.common.pruning.operations import TransposeConvolutionPruningOp
from nncf.common.pruning.utils import PruningOperationsMetatypeRegistry
from nncf.tensorflow.graph.metatypes import keras_layers as layer_metatypes
from nncf.tensorflow.graph.metatypes import tf_ops as op_metatypes
from nncf.tensorflow.graph.pattern_operations import ELEMENTWISE_OPERATIONS
from nncf.tensorflow.graph.pattern_operations import KERAS_ACTIVATIONS_OPERATIONS
from nncf.tensorflow.graph.pattern_operations import TF_ACTIVATIONS_OPERATIONS

TF_PRUNING_OPERATOR_METATYPES = PruningOperationsMetatypeRegistry("operator_metatypes")


def _get_types(operations_dict: Dict) -> List[str]:
    return operations_dict["type"]


@TF_PRUNING_OPERATOR_METATYPES.register("model_input")
class TFInputPruningOp(InputPruningOp):
    additional_types = ["InputLayer", NNCFGraphNodeType.INPUT_NODE]


@TF_PRUNING_OPERATOR_METATYPES.register("model_output")
class TFOutputPruningOp(OutputPruningOp):
    additional_types = [NNCFGraphNodeType.OUTPUT_NODE]


@TF_PRUNING_OPERATOR_METATYPES.register("identity_mask_propagation")
class TFIdentityMaskForwardPruningOp(IdentityMaskForwardPruningOp):
    additional_types = (
        _get_types(KERAS_ACTIVATIONS_OPERATIONS)
        + _get_types(TF_ACTIVATIONS_OPERATIONS)
        + layer_metatypes.TFAveragePooling2DLayerMetatype.get_all_aliases()
        + layer_metatypes.TFGlobalAveragePooling2DLayerMetatype.get_all_aliases()
        + layer_metatypes.TFMaxPooling2DLayerMetatype.get_all_aliases()
        + layer_metatypes.TFGlobalMaxPooling2DLayerMetatype.get_all_aliases()
        + layer_metatypes.TFZeroPadding2DLayerMetatype.get_all_aliases()
        + layer_metatypes.TFUpSampling2DLayerMetatype.get_all_aliases()
        + layer_metatypes.TFAveragePooling3DLayerMetatype.get_all_aliases()
        + layer_metatypes.TFGlobalAveragePooling3DLayerMetatype.get_all_aliases()
        + layer_metatypes.TFMaxPooling3DLayerMetatype.get_all_aliases()
        + layer_metatypes.TFGlobalMaxPooling3DLayerMetatype.get_all_aliases()
        + layer_metatypes.TFZeroPadding3DLayerMetatype.get_all_aliases()
        + layer_metatypes.TFUpSampling3DLayerMetatype.get_all_aliases()
        + layer_metatypes.TFDropoutLayerMetatype.get_all_aliases()
        + op_metatypes.TFIdentityOpMetatype.get_all_aliases()
        + op_metatypes.TFPadOpMetatype.get_all_aliases()
    )


@TF_PRUNING_OPERATOR_METATYPES.register("convolution")
class TFConvolutionPruningOp(ConvolutionPruningOp):
    additional_types = ["Conv1D", "Conv2D", "Conv3D", "DepthwiseConv2D"]


@TF_PRUNING_OPERATOR_METATYPES.register("transpose_convolution")
class TFTransposeConvolutionPruningOp(TransposeConvolutionPruningOp):
    additional_types = ["Conv1DTranspose", "Conv2DTranspose", "Conv3DTranspose"]


@TF_PRUNING_OPERATOR_METATYPES.register("linear")
class TFLinearPruningOp(LinearPruningOp):
    additional_types = (
        layer_metatypes.TFDenseLayerMetatype.get_all_aliases() + op_metatypes.TFMatMulOpMetatype.get_all_aliases()
    )


@TF_PRUNING_OPERATOR_METATYPES.register("batch_norm")
class TFBatchNormPruningOp(BatchNormPruningOp):
    additional_types = ["BatchNormalization", "SyncBatchNormalization"]


@TF_PRUNING_OPERATOR_METATYPES.register("elementwise")
class TFElementwisePruningOp(ElementwisePruningOp):
    additional_types = _get_types(ELEMENTWISE_OPERATIONS)


@TF_PRUNING_OPERATOR_METATYPES.register("reshape")
class TFReshapeOps(ReshapePruningOp):
    additional_types = op_metatypes.TFReshapeOpMetatype.get_all_aliases()


@TF_PRUNING_OPERATOR_METATYPES.register("flatten")
class TFFlattenOps(FlattenPruningOp):
    additional_types = layer_metatypes.TFFlattenLayerMetatype.get_all_aliases()


@TF_PRUNING_OPERATOR_METATYPES.register("stop_propagation_ops")
class TFStopMaskForwardPruningOp(StopMaskForwardPruningOp):
    additional_types = []


@TF_PRUNING_OPERATOR_METATYPES.register("concat")
class TFConcatPruningOp(ConcatPruningOp):
    additional_types = (
        layer_metatypes.TFConcatenateLayerMetatype.get_all_aliases() + op_metatypes.TFConcatOpMetatype.get_all_aliases()
    )


@TF_PRUNING_OPERATOR_METATYPES.register("split")
class TFSplitPruningOp(SplitPruningOp):
    additional_types = op_metatypes.TFSplitOpMetatype.get_all_aliases()



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.config.definitions import ONLINE_DOCS_ROOT
from nncf.config.definitions import RB_SPARSITY_ALGO_NAME_IN_CONFIG
from nncf.config.schemata.basic import NUMBER
from nncf.config.schemata.basic import with_attributes
from nncf.config.schemata.common.compression import BASIC_COMPRESSION_ALGO_SCHEMA
from nncf.config.schemata.common.compression import COMPRESSION_LR_MULTIPLIER_PROPERTY
from nncf.config.schemata.common.sparsity import COMMON_SPARSITY_PARAM_PROPERTIES
from nncf.config.schemata.common.targeting import SCOPING_PROPERTIES
from nncf.config.schemata.defaults import SPARSITY_INIT

RB_SPARSITY_SCHEMA = {
    **BASIC_COMPRESSION_ALGO_SCHEMA,
    "description": f"Applies sparsity on top of the current model. Each weight tensor value will be either kept as-is, "
    f"or set to 0 based on its importance as determined by the regularization-based sparsity algorithm. "
    f"For large sparsity levels, this will improve performance on "
    f"hardware that can profit from it. "
    f"See [Sparsity.md]"
    f"({ONLINE_DOCS_ROOT}"
    f"/docs/compression_algorithms/Sparsity.md#rb-sparsity) and the rest of this schema for "
    f"more details and parameters.",
    "properties": {
        "algorithm": {"const": RB_SPARSITY_ALGO_NAME_IN_CONFIG},
        "sparsity_init": with_attributes(
            NUMBER, description="Initial value of the sparsity level applied to the model", default=SPARSITY_INIT
        ),
        "params": {"type": "object", "properties": COMMON_SPARSITY_PARAM_PROPERTIES, "additionalProperties": False},
        **SCOPING_PROPERTIES,
        **COMPRESSION_LR_MULTIPLIER_PROPERTY,
    },
    "additionalProperties": False,
}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import tensorflow as tf

from nncf.common.accuracy_aware_training import create_accuracy_aware_training_loop
from nncf.config.structures import ModelEvaluationArgs
from nncf.tensorflow import tf_internals


def accuracy_aware_fit(
    cls_instance,
    train_dataset,
    compression_ctrl,
    nncf_config,
    callbacks,
    initial_epoch,
    uncompressed_model_accuracy,
    steps_per_epoch=None,
    batch_size=None,
    tensorboard_writer=None,
    log_dir=None,
    validation_data=None,
    validation_steps=None,
    result_dict_to_val_metric_fn=None,
    **kwargs,
):
    if result_dict_to_val_metric_fn is None:
        result_dict_to_val_metric_fn = lambda metric: metric

    with cls_instance.distribute_strategy.scope(), tf_internals.keras_engine.training_utils.RespectCompiledTrainableState(  # noqa: E501
        cls_instance
    ):
        data_handler = tf_internals.keras_engine.data_adapter.DataHandler(
            x=train_dataset,
            y=None,
            sample_weight=None,
            batch_size=batch_size,
            steps_per_epoch=steps_per_epoch,
            initial_epoch=initial_epoch,
            epochs=1,
            shuffle=True,
            class_weight=None,
            max_queue_size=10,
            workers=1,
            use_multiprocessing=False,
            model=cls_instance,
            steps_per_execution=cls_instance._steps_per_execution,
        )

        if not isinstance(callbacks, tf.keras.callbacks.CallbackList):
            callbacks = tf.keras.callbacks.CallbackList(
                callbacks,
                add_history=True,
                model=cls_instance,
                epochs=1,
                verbose=1,
                add_progbar=True,
                steps=data_handler.inferred_steps,
            )

    def train_epoch_fn(compression_ctrl, model, epoch, **kwargs):
        model.reset_metrics()

        if model.train_function is None:
            model.train_function = model.make_train_function()
        _, iterator = next(data_handler.enumerate_epochs())

        callbacks.on_epoch_begin(epoch)
        with data_handler.catch_stop_iteration():
            for step in data_handler.steps():
                with tf.profiler.experimental.Trace("train", epoch_num=epoch, step_num=step, batch_size=None, _r=1):
                    callbacks.on_train_batch_begin(step)
                    tmp_logs = model.train_function(iterator)
                    if data_handler.should_sync:
                        tf_internals.eager_context.async_wait()
                    logs = tmp_logs
                    end_step = step + data_handler.step_increment
                    callbacks.on_train_batch_end(end_step, logs)
                    if model.stop_training:
                        break

        if logs is None:
            raise ValueError("Expect x to be a non-empty array or dataset.")
        epoch_logs = copy.copy(logs)
        callbacks.on_epoch_end(epoch, epoch_logs)

    if validation_data is None:
        validation_data = train_dataset

    def validate_fn(model, epoch=None):
        val_x, val_y, val_sample_weight = tf.keras.utils.unpack_x_y_sample_weight(validation_data)
        val_logs = model.evaluate(
            x=val_x,
            y=val_y,
            sample_weight=val_sample_weight,
            batch_size=None,
            steps=validation_steps,
            callbacks=callbacks,
            return_dict=True,
        )
        return result_dict_to_val_metric_fn(val_logs)

    current_optimizer = copy.copy(compression_ctrl.model.optimizer)

    def configure_optimizers_fn():
        optimizer = copy.copy(current_optimizer)
        return optimizer, None

    nncf_config.register_extra_structs([ModelEvaluationArgs(eval_fn=validate_fn)])
    callbacks.on_train_begin()
    acc_aware_training_loop = create_accuracy_aware_training_loop(
        nncf_config, compression_ctrl, uncompressed_model_accuracy
    )
    cls_instance = acc_aware_training_loop.run(
        cls_instance,
        train_epoch_fn=train_epoch_fn,
        validate_fn=validate_fn,
        configure_optimizers_fn=configure_optimizers_fn,
        tensorboard_writer=tensorboard_writer,
        log_dir=log_dir,
    )
    callbacks.on_train_end()

    return acc_aware_training_loop.statistics



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.initialization.batchnorm_adaptation import BatchnormAdaptationAlgorithmImpl
from nncf.torch.initialization import DataLoaderBNAdaptationRunner


class PTBatchnormAdaptationAlgorithmImpl(BatchnormAdaptationAlgorithmImpl):
    """
    Implementation of the batch-norm statistics adaptation algorithm for the PyTorch.
    """

    def run(self, model) -> None:
        """
        Runs the batch-norm statistics adaptation algorithm.

        :param model: A model for which the algorithm will be applied.
        """
        bn_adaptation_runner = DataLoaderBNAdaptationRunner(model, self._device)
        bn_adaptation_runner.run(self._data_loader, self._num_bn_adaptation_steps)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import numbers
from typing import Dict, Tuple, Union

import torch
import torch.nn.functional as F
from torch import nn
from torch.nn import init
from torch.nn.utils.rnn import PackedSequence
from torch.nn.utils.weight_norm import WeightNorm

import nncf
from nncf import nncf_logger
from nncf.common.graph.layer_attributes import GenericWeightedLayerAttributes
from nncf.common.utils.api_marker import api
from nncf.common.utils.registry import Registry
from nncf.torch.checkpoint_loading import OPTIONAL_PARAMETERS_REGISTRY
from nncf.torch.dynamic_graph.context import forward_nncf_trace
from nncf.torch.layer_utils import _NNCFModuleMixin
from nncf.torch.utils import no_jit_trace


def dict_update(src: Dict, dst: Dict, recursive: bool = True):
    for name, value in src.items():
        if recursive and name in dst and isinstance(value, dict):
            dict_update(value, dst[name], recursive)
        else:
            dst[name] = value


def maybe_reapply_weight_norm(src: torch.nn.Module, dst: torch.nn.Module) -> torch.nn.Module:
    for k, hook in src._forward_pre_hooks.items():
        if isinstance(hook, WeightNorm):
            # The code below presumes that the `hook` object does not
            # contain internal references to the module it was set up on
            # (i.e. to the `src`) and takes the module to act on as a parameter.
            # This is the case for the `WeightNorm` hook.
            hook.remove(dst)
            del dst._forward_pre_hooks[k]
            name = hook.name
            dim = hook.dim
            WeightNorm.apply(dst, name=name, dim=dim)
    return dst


def align_module_internals(src: torch.nn.Module, dst: torch.nn.Module) -> torch.nn.Module:
    dict_update(src.__dict__, dst.__dict__)
    dst = maybe_reapply_weight_norm(src, dst)
    return dst


class NNCFConv1d(_NNCFModuleMixin, nn.Conv1d):
    op_func_name = "conv1d"

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.Conv1d.__name__
        nncf_conv = NNCFConv1d(
            module.in_channels,
            module.out_channels,
            module.kernel_size,
            module.stride,
            module.padding,
            module.dilation,
            module.groups,
            hasattr(module, "bias"),
        )
        nncf_conv = align_module_internals(module, nncf_conv)
        return nncf_conv


class NNCFConvTranspose1d(_NNCFModuleMixin, nn.ConvTranspose1d):
    op_func_name = "conv_transpose1d"
    target_weight_dim_for_compression = 1

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.ConvTranspose1d.__name__
        args = [
            module.in_channels,
            module.out_channels,
            module.kernel_size,
            module.stride,
            module.padding,
            module.output_padding,
            module.groups,
            hasattr(module, "bias"),
            module.dilation,
        ]
        if hasattr(module, "padding_mode"):
            args.append(module.padding_mode)
        nncf_conv_transpose1d = NNCFConvTranspose1d(*args)
        nncf_conv_transpose1d = align_module_internals(module, nncf_conv_transpose1d)
        return nncf_conv_transpose1d


NNCF_PADDING_VALUE_ATTR_NAME = "nncf_padding_value"
OPTIONAL_PARAMETERS_REGISTRY.register(NNCF_PADDING_VALUE_ATTR_NAME)


class NNCFConv2d(_NNCFModuleMixin, nn.Conv2d):
    op_func_name = "conv2d"

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
    ):
        super().__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias, padding_mode)
        self.register_buffer(NNCF_PADDING_VALUE_ATTR_NAME, torch.zeros([1]))

    def get_padding_value_ref(self):
        return getattr(self, NNCF_PADDING_VALUE_ATTR_NAME)

    def _set_padding_value(self, value):
        setattr(self, NNCF_PADDING_VALUE_ATTR_NAME, value)

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.Conv2d.__name__
        nncf_conv = NNCFConv2d(
            module.in_channels,
            module.out_channels,
            module.kernel_size,
            module.stride,
            module.padding,
            module.dilation,
            module.groups,
            hasattr(module, "bias"),
        )
        nncf_conv = align_module_internals(module, nncf_conv)
        return nncf_conv

    # override class attribute of _NNCFModuleMixin
    def _custom_forward_fn(self, input_):
        proxy_padding_value = getattr(self, NNCF_PADDING_VALUE_ATTR_NAME)  # hack to get value from ProxyModule
        proxy_weight = self.weight
        proxy_bias = self.bias
        proxy_padding = self.padding
        proxy_num_groups = self.groups
        return self._conv_forward_proxy(
            input_, proxy_weight, proxy_bias, proxy_padding_value, proxy_padding, proxy_num_groups
        )

    def _conv_forward_proxy(self, input_, weight, bias, padding_value, padding, num_groups):
        with no_jit_trace():
            padding_val = padding_value.item()
        self.get_padding_value_ref().data.fill_(padding_val)
        self.groups = num_groups

        def _reverse_repeat_tuple(t, n):
            r"""Reverse the order of `t` and repeat each element for `n` times.

            This can be used to translate padding arg used by Conv and Pooling modules
            to the ones used by `F.pad`.
            """
            return tuple(x for x in reversed(t) for _ in range(n))

        reversed_padding_repeated_twice = _reverse_repeat_tuple(self.padding, 2)
        if self.padding_mode != "zeros":
            return F.conv2d(
                F.pad(input_, reversed_padding_repeated_twice, mode=self.padding_mode, value=padding_val),
                weight,
                bias,
                self.stride,
                (0, 0),
                self.dilation,
                self.groups,
            )
        if padding_val == 0:
            return F.conv2d(input_, weight, bias, self.stride, padding, self.dilation, self.groups)
        return F.conv2d(
            F.pad(input_, reversed_padding_repeated_twice, value=padding_val),
            weight,
            bias,
            self.stride,
            (0, 0),
            self.dilation,
            self.groups,
        )


class NNCFLinear(_NNCFModuleMixin, nn.Linear):
    op_func_name = "linear"

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.Linear.__name__

        nncf_linear = NNCFLinear(module.in_features, module.out_features, hasattr(module, "bias"))
        nncf_linear = align_module_internals(module, nncf_linear)
        return nncf_linear


class NNCFBatchNorm1d(_NNCFModuleMixin, nn.BatchNorm1d):
    op_func_name = "batch_norm"
    ignored_algorithms = ["magnitude_sparsity", "rb_sparsity", "const_sparsity", "quantization"]

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.BatchNorm1d.__name__

        nncf_bn = NNCFBatchNorm1d(module.num_features)
        nncf_bn = align_module_internals(module, nncf_bn)
        return nncf_bn


class NNCFBatchNorm2d(_NNCFModuleMixin, nn.BatchNorm2d):
    op_func_name = "batch_norm"
    ignored_algorithms = ["magnitude_sparsity", "rb_sparsity", "const_sparsity", "quantization"]

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.BatchNorm2d.__name__

        nncf_bn = NNCFBatchNorm2d(module.num_features)
        nncf_bn = align_module_internals(module, nncf_bn)
        return nncf_bn


class NNCFBatchNorm3d(_NNCFModuleMixin, nn.BatchNorm3d):
    op_func_name = "batch_norm"
    ignored_algorithms = ["magnitude_sparsity", "rb_sparsity", "const_sparsity", "quantization"]

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.BatchNorm3d.__name__

        nncf_bn = NNCFBatchNorm3d(module.num_features)
        nncf_bn = align_module_internals(module, nncf_bn)
        return nncf_bn


class NNCFGroupNorm(_NNCFModuleMixin, nn.GroupNorm):
    op_func_name = "group_norm"
    ignored_algorithms = ["magnitude_sparsity", "rb_sparsity", "const_sparsity", "quantization"]

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.GroupNorm.__name__

        nncf_bn = NNCFGroupNorm(
            num_groups=module.num_groups, num_channels=module.num_channels, eps=module.eps, affine=module.affine
        )
        nncf_bn = align_module_internals(module, nncf_bn)
        return nncf_bn


class NNCFLayerNorm(_NNCFModuleMixin, nn.LayerNorm):
    op_func_name = "layer_norm"
    ignored_algorithms = ["magnitude_sparsity", "rb_sparsity", "const_sparsity", "quantization"]

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.LayerNorm.__name__

        nncf_ln = NNCFLayerNorm(module.normalized_shape, hasattr(module, "eps"))
        nncf_ln = align_module_internals(module, nncf_ln)
        return nncf_ln


class NNCFConvTranspose2d(_NNCFModuleMixin, nn.ConvTranspose2d):
    op_func_name = "conv_transpose2d"
    target_weight_dim_for_compression = 1

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.ConvTranspose2d.__name__
        args = [
            module.in_channels,
            module.out_channels,
            module.kernel_size,
            module.stride,
            module.padding,
            module.output_padding,
            module.groups,
            hasattr(module, "bias"),
            module.dilation,
        ]
        if hasattr(module, "padding_mode"):
            args.append(module.padding_mode)
        nncf_conv_transpose2d = NNCFConvTranspose2d(*args)
        nncf_conv_transpose2d = align_module_internals(module, nncf_conv_transpose2d)
        return nncf_conv_transpose2d


class NNCFConv3d(_NNCFModuleMixin, nn.Conv3d):
    op_func_name = "conv3d"

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.Conv3d.__name__

        nncf_conv3d = NNCFConv3d(
            module.in_channels,
            module.out_channels,
            module.kernel_size,
            module.stride,
            module.padding,
            module.dilation,
            module.groups,
            hasattr(module, "bias"),
        )
        nncf_conv3d = align_module_internals(module, nncf_conv3d)
        return nncf_conv3d


class NNCFConvTranspose3d(_NNCFModuleMixin, nn.ConvTranspose3d):
    op_func_name = "conv_transpose3d"
    target_weight_dim_for_compression = 1

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.ConvTranspose3d.__name__
        args = [
            module.in_channels,
            module.out_channels,
            module.kernel_size,
            module.stride,
            module.padding,
            module.output_padding,
            module.groups,
            hasattr(module, "bias"),
            module.dilation,
        ]
        if hasattr(module, "padding_mode"):
            args.append(module.padding_mode)
        nncf_conv_transpose3d = NNCFConvTranspose3d(*args)
        nncf_conv_transpose3d = align_module_internals(module, nncf_conv_transpose3d)
        return nncf_conv_transpose3d


class NNCFEmbedding(_NNCFModuleMixin, nn.Embedding):
    op_func_name = "embedding"
    target_weight_dim_for_compression = 0

    # Note that this does not require activation quantization because it's basically a lookup.
    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.Embedding.__name__

        args = [
            module.num_embeddings,
            module.embedding_dim,
            module.padding_idx,
            module.max_norm,
            module.norm_type,
            module.scale_grad_by_freq,
            module.sparse,
            module.weight,
        ]
        nncf_embedding = NNCFEmbedding(*args)
        nncf_embedding = align_module_internals(module, nncf_embedding)
        return nncf_embedding


class NNCFEmbeddingBag(_NNCFModuleMixin, nn.EmbeddingBag):
    op_func_name = "embedding_bag"

    @staticmethod
    def from_module(module):
        assert module.__class__.__name__ == nn.EmbeddingBag.__name__

        args = [
            module.num_embeddings,
            module.embedding_dim,
            module.max_norm,
            module.norm_type,
            module.scale_grad_by_freq,
            module.mode,
            module.sparse,
            module.weight,
            module.include_last_offset,
        ]
        nncf_embedding_bag = NNCFEmbeddingBag(*args)
        nncf_embedding_bag = align_module_internals(module, nncf_embedding_bag)
        return nncf_embedding_bag


NNCF_MODULES_DICT = {
    NNCFConv1d: nn.Conv1d,
    NNCFConv2d: nn.Conv2d,
    NNCFConv3d: nn.Conv3d,
    NNCFLinear: nn.Linear,
    NNCFBatchNorm1d: nn.BatchNorm1d,
    NNCFBatchNorm2d: nn.BatchNorm2d,
    NNCFBatchNorm3d: nn.BatchNorm3d,
    NNCFGroupNorm: nn.GroupNorm,
    NNCFLayerNorm: nn.LayerNorm,
    NNCFConvTranspose1d: nn.ConvTranspose1d,
    NNCFConvTranspose2d: nn.ConvTranspose2d,
    NNCFConvTranspose3d: nn.ConvTranspose3d,
    NNCFEmbedding: nn.Embedding,
    NNCFEmbeddingBag: nn.EmbeddingBag,
}

NNCF_MODULES_MAP = {k.__name__: v.__name__ for k, v in NNCF_MODULES_DICT.items()}
NNCF_MODULES = list(NNCF_MODULES_MAP.keys())
NNCF_MODULES_OP_NAMES = [k.op_func_name for k, _ in NNCF_MODULES_DICT.items()]

NNCF_CONV_MODULES_DICT = {
    NNCFConv1d: nn.Conv1d,
    NNCFConv2d: nn.Conv2d,
    NNCFConv3d: nn.Conv3d,
}
NNCF_DECONV_MODULES_DICT = {
    NNCFConvTranspose1d: nn.ConvTranspose1d,
    NNCFConvTranspose2d: nn.ConvTranspose2d,
    NNCFConvTranspose3d: nn.ConvTranspose3d,
}
NNCF_CONV_MODULES_MAP = {k.__name__: v.__name__ for k, v in NNCF_CONV_MODULES_DICT.items()}
NNCF_CONV_MODULES = list(NNCF_CONV_MODULES_MAP.keys())
NNCF_GENERAL_CONV_MODULES_DICT = dict(NNCF_CONV_MODULES_DICT)
NNCF_GENERAL_CONV_MODULES_DICT.update(NNCF_DECONV_MODULES_DICT)

NNCF_LINEAR_MODULES_DICT = {NNCFLinear: nn.Linear}
NNCF_MODULES_OP_NAMES = [k.op_func_name for k, _ in NNCF_MODULES_DICT.items()]

NNCF_PRUNING_MODULES_DICT = {
    NNCFLinear: nn.Linear,
    NNCFConv1d: nn.Conv1d,
    NNCFConv2d: nn.Conv2d,
    NNCFConv3d: nn.Conv3d,
    NNCFConvTranspose1d: nn.ConvTranspose1d,
    NNCFConvTranspose2d: nn.ConvTranspose2d,
    NNCFConvTranspose3d: nn.ConvTranspose3d,
}
NNCF_PRUNING_MODULES_MAP = {k.__name__: v.__name__ for k, v in NNCF_CONV_MODULES_DICT.items()}
NNCF_PRUNING_MODULES = list(NNCF_CONV_MODULES_MAP.keys())

UNWRAPPED_USER_MODULES = Registry("user_modules")
NNCF_WRAPPED_USER_MODULES_DICT = {}


@api(canonical_alias="nncf.torch.register_module")
def register_module(
    *quantizable_field_names: str, ignored_algorithms: list = None, target_weight_dim_for_compression: int = 0
):
    # quantizable_field_names will work for `weight` attributes only. Should later extend to registering
    # customly named attributes if it becomes necessary
    def wrap(cls):
        UNWRAPPED_USER_MODULES.registry_dict[cls.__name__] = cls
        nncf_wrapped_module_class_name = "NNCFUser{}".format(cls.__name__)
        NNCF_WRAPPED_USER_MODULES_DICT[cls] = type(nncf_wrapped_module_class_name, (_NNCFModuleMixin, cls), {})
        get_base_attributes_fn = lambda self: GenericWeightedLayerAttributes(
            self.weight.requires_grad, self.weight.shape
        )
        setattr(NNCF_WRAPPED_USER_MODULES_DICT[cls], "get_weight_shape", get_base_attributes_fn)
        if ignored_algorithms:
            setattr(NNCF_WRAPPED_USER_MODULES_DICT[cls], "ignored_algorithms", ignored_algorithms)

        setattr(
            NNCF_WRAPPED_USER_MODULES_DICT[cls], "target_weight_dim_for_compression", target_weight_dim_for_compression
        )
        return cls

    return wrap


def add_nncf_functionality_to_user_module(module: torch.nn.Module):
    user_class = module.__class__
    assert user_class.__name__ in UNWRAPPED_USER_MODULES.registry_dict
    module.__class__ = NNCF_WRAPPED_USER_MODULES_DICT[user_class]
    _NNCFModuleMixin.add_mixin_fields(module)
    return module


class RNNCellBaseNNCF(nn.Module):
    __constants__ = ["input_size", "hidden_size", "bias"]

    def __init__(self, input_size, hidden_size, bias, num_chunks):
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        linear_ih = nn.Linear(input_size, num_chunks * hidden_size, self.bias)
        linear_hh = nn.Linear(hidden_size, num_chunks * hidden_size, self.bias)
        self.weight_ih = linear_ih.weight
        self.weight_hh = linear_hh.weight
        self.bias_ih = linear_ih.bias
        self.bias_hh = linear_hh.bias
        self.linear_list = [linear_ih, linear_hh]
        self.reset_parameters()

    def extra_repr(self):
        s = "{input_size}, {hidden_size}"
        if "bias" in self.__dict__ and self.bias is not True:
            s += ", bias={bias}"
        if "nonlinearity" in self.__dict__ and self.nonlinearity != "tanh":
            s += ", nonlinearity={nonlinearity}"
        return s.format(**self.__dict__)

    def check_forward_input(self, input_):
        if input_.size(1) != self.input_size:
            raise nncf.ValidationError(
                "input_ has inconsistent input_size: got {}, expected {}".format(input_.size(1), self.input_size)
            )

    def check_forward_hidden(self, input_: torch.Tensor, hx: torch.Tensor, hidden_label: str = ""):
        if input_.size(0) != hx.size(0):
            raise nncf.ValidationError(
                "Input batch size {} doesn't match hidden{} batch size {}".format(
                    input_.size(0), hidden_label, hx.size(0)
                )
            )

        if hx.size(1) != self.hidden_size:
            raise nncf.ValidationError(
                "hidden{} has inconsistent hidden_size: got {}, expected {}".format(
                    hidden_label, hx.size(1), self.hidden_size
                )
            )

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.hidden_size)
        for weight in self.parameters():
            init.uniform_(weight, -stdv, stdv)

    def forward(self, input_, hidden):
        raise NotImplementedError


ITERATION_MODULES = Registry("iteration_modules")


@ITERATION_MODULES.register()
class LSTMCellForwardNNCF(nn.Module):
    def __init__(self, input_linear, hidden_linear):
        super().__init__()
        self.input_linear = input_linear
        self.hidden_linear = hidden_linear

    def forward(self, input_: torch.Tensor, hidden: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        hx, cx = hidden
        gates = self.input_linear(input_) + self.hidden_linear(hx)

        ingate, forgetgate, cellgate, outgate = gates.chunk(4, 1)

        ingate = F.sigmoid(ingate)
        forgetgate = F.sigmoid(forgetgate)
        cellgate = F.tanh(cellgate)
        outgate = F.sigmoid(outgate)

        cy = (forgetgate * cx) + (ingate * cellgate)
        hy = outgate * F.tanh(cy)
        return hy, cy


class LSTMCellNNCF(RNNCellBaseNNCF):
    def __init__(self, input_size=1, hidden_size=1, bias=True):
        super().__init__(input_size, hidden_size, bias, num_chunks=4)
        self.cell = LSTMCellForwardNNCF(self.linear_list[0], self.linear_list[1])

    def forward(self, input_, hidden=None):
        self.check_forward_input(input_)
        if hidden is None:
            zeros = torch.zeros(input_.size(0), self.hidden_size, dtype=input_.dtype, device=input_.device)
            hidden = (zeros, zeros)
        self.check_forward_hidden(input_, hidden[0], "[0]")
        self.check_forward_hidden(input_, hidden[1], "[1]")

        return self.cell(input_, hidden)


@ITERATION_MODULES.register()
class StackedRNN(nn.Module):
    class StackedRNNResetPoint(nn.Module):
        """
        Intentionally wrap concat, which is called inside nested loops, as a separate module.
        It allows not to add new node to nncf graph on each iteration of the loops.
        """

        def forward(self, all_output, input_):
            input_ = torch.cat(all_output, input_.dim() - 1)
            return input_

    def __init__(self, inners, num_layers, lstm=False, dropout=0):
        super().__init__()
        self.lstm = lstm
        self.num_layers = num_layers
        self.num_directions = int(len(inners) / num_layers)
        self.inners = nn.ModuleList(inners)
        self.total_layers = self.num_layers * self.num_directions
        self.dropout = dropout

    def forward(self, input_, hidden, batch_sizes):
        next_hidden = []

        if self.lstm:
            hidden = list(zip(*hidden))

        for i in range(self.num_layers):
            all_output = []
            for j in range(self.num_directions):
                k = i * self.num_directions + j
                hy, output = self.inners[k](input_, hidden[k], batch_sizes)
                next_hidden.append(hy)
                all_output.append(output)

            input_ = self.StackedRNNResetPoint()(all_output, input_)
            if self.dropout != 0 and i < self.num_layers - 1:
                input_ = F.dropout(input_, p=self.dropout, training=self.training, inplace=False)

        if self.lstm:
            next_h, next_c = zip(*next_hidden)
            next_hidden = (
                torch.cat(next_h, 0).view(self.total_layers, *next_h[0].size()),
                torch.cat(next_c, 0).view(self.total_layers, *next_c[0].size()),
            )
        else:
            next_hidden = torch.cat(next_hidden, 0).view(self.total_layers, *next_hidden[0].size())
        return next_hidden, input_


@ITERATION_MODULES.register()
class Recurrent(nn.Module):
    def __init__(self, cell, reverse=False):
        super().__init__()
        self.reverse = reverse
        self.cell = cell

    def forward(self, input_, hidden, batch_sizes=None):
        output = []
        steps = range(input_.size(0) - 1, -1, -1) if self.reverse else range(input_.size(0))
        for i in steps:
            with forward_nncf_trace():
                hidden_input = input_[i]
            hidden = self.cell(hidden_input, hidden)
            output.append(hidden[0] if isinstance(hidden, tuple) else hidden)

        if self.reverse:
            output.reverse()
        output = torch.cat(output, 0).view(input_.size(0), *output[0].size())

        return hidden, output


def variable_recurrent_factory():
    def factory(cell, reverse=False):
        if reverse:
            return VariableRecurrentReverse(cell)
        return VariableRecurrent(cell)

    return factory


@ITERATION_MODULES.register()
class VariableRecurrent(nn.Module):
    def __init__(self, cell):
        super().__init__()
        self.cell = cell

    def forward(self, input_, hidden, batch_sizes):
        output = []
        input_offset = 0
        last_batch_size = batch_sizes[0]
        hiddens = []
        flat_hidden = not isinstance(hidden, tuple)
        if flat_hidden:
            hidden = (hidden,)
        with forward_nncf_trace():
            batch_size_elements = [b for b in batch_sizes]
        for batch_size in batch_size_elements:
            step_input = input_[input_offset : input_offset + batch_size]
            input_offset += batch_size

            bs_decrease = last_batch_size - batch_size
            if bs_decrease > 0:
                hidden_len = len(hidden)
                hidden_offset_elts = []
                hidden_offset_elts_reversed = []
                for i in range(hidden_len):
                    with forward_nncf_trace():
                        hidden_offset_elts.append(hidden[i][-bs_decrease:])
                        hidden_offset_elts_reversed.append(hidden[i][:-bs_decrease])

                hiddens.append(tuple(hidden_offset_elts))
                hidden = tuple(hidden_offset_elts_reversed)
            last_batch_size = batch_size

            if flat_hidden:
                hidden = (self.cell(step_input, hidden[0]),)
            else:
                hidden = self.cell(step_input, hidden)

            output.append(hidden[0])
        hiddens.append(hidden)
        hiddens.reverse()

        hidden = tuple(torch.cat(h, 0) for h in zip(*hiddens))
        assert hidden[0].size(0) == batch_sizes[0]
        if flat_hidden:
            hidden = hidden[0]
        output = torch.cat(output, 0)
        return hidden, output


@ITERATION_MODULES.register()
class VariableRecurrentReverse(nn.Module):
    def __init__(self, cell):
        super().__init__()
        self.cell = cell

    def forward(self, input_, hidden, batch_sizes):
        output = []
        input_offset = input_.size(0)
        last_batch_size = batch_sizes[-1]
        initial_hidden = hidden
        flat_hidden = not isinstance(hidden, tuple)
        if flat_hidden:
            hidden = (hidden,)
            initial_hidden = (initial_hidden,)
        hidden = tuple(h[: batch_sizes[-1]] for h in hidden)
        for batch_size in reversed(batch_sizes):
            inc = batch_size - last_batch_size
            hidden = self.ReverseResetPoint()(batch_size, hidden, inc, initial_hidden, last_batch_size)
            last_batch_size = batch_size
            step_input = input_[input_offset - batch_size : input_offset]
            input_offset -= batch_size

            if flat_hidden:
                hidden = (self.cell(step_input, hidden[0]),)
            else:
                hidden = self.cell(step_input, hidden)
            output.append(hidden[0])

        output.reverse()
        output = torch.cat(output, 0)
        if flat_hidden:
            hidden = hidden[0]
        return hidden, output

    @ITERATION_MODULES.register()
    class ReverseResetPoint(nn.Module):
        """
        Intentionally wrap concat undef if condition as a separate module
        to prevent adding new node to nncf graph on each iteration
        """

        def forward(self, batch_size, hidden, inc, initial_hidden, last_batch_size):
            if inc > 0:
                hidden = tuple(
                    torch.cat((h, ih[last_batch_size:batch_size]), 0) for h, ih in zip(hidden, initial_hidden)
                )
            return hidden


class NNCF_RNN(nn.Module):
    """Common class for RNN modules. Currently, LSTM is supported only"""

    def __init__(
        self,
        mode="LSTM",
        input_size=1,
        hidden_size=1,
        num_layers=1,
        batch_first=False,
        dropout=0,
        bidirectional=False,
        bias=True,
    ):
        super().__init__()
        self.mode = mode
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bias = bias
        self.batch_first = batch_first
        self.dropout = dropout
        self.bidirectional = bidirectional
        self.num_directions = 2 if bidirectional else 1

        if not isinstance(dropout, numbers.Number) or not 0 <= dropout <= 1 or isinstance(dropout, bool):
            raise ValueError(
                "dropout should be a number in range [0, 1] "
                "representing the probability of an element being "
                "zeroed"
            )
        if dropout > 0 and num_layers == 1:
            nncf_logger.debug(
                f"dropout option adds dropout after all but last recurrent layer, "
                f"so non-zero dropout expects num_layers greater than 1, "
                f"but got dropout={dropout} and num_layers={num_layers}"
            )

        if mode == "LSTM":
            gate_size = 4 * hidden_size
            self.cell_type = LSTMCellForwardNNCF
        else:
            # elif mode == 'GRU':
            #     gate_size = 3 * hidden_size
            # elif mode == 'RNN_TANH':
            #     gate_size = hidden_size
            # elif mode == 'RNN_RELU':
            #     gate_size = hidden_size
            # else:
            raise ValueError("Unrecognized RNN mode: " + mode)

        self._all_weights = []
        self.cells = []
        for layer in range(num_layers):
            for direction in range(self.num_directions):
                layer_input_size = input_size if layer == 0 else hidden_size * self.num_directions
                linear_ih = nn.Linear(layer_input_size, gate_size, bias)
                linear_hh = nn.Linear(hidden_size, gate_size, bias)
                self.cells.append(self.cell_type(linear_ih, linear_hh))
                params = (linear_ih.weight, linear_hh.weight, linear_ih.bias, linear_hh.bias)
                suffix = "_reverse" if direction == 1 else ""
                weight_names = ["weight_ih_l{}{}", "weight_hh_l{}{}"]
                if bias:
                    weight_names += ["bias_ih_l{}{}", "bias_hh_l{}{}"]
                weight_names = [x.format(layer, suffix) for x in weight_names]
                for name, param in zip(weight_names, params):
                    setattr(self, name, param)
                self._all_weights.append(weight_names)

        self.reset_parameters()
        self.variable_length = True
        self.rnn_impl = self.get_rnn_impl(self.variable_length, self.cells)

    def get_rnn_impl(self, variable_length, cells):
        if variable_length:
            rec_factory = variable_recurrent_factory()
        else:
            rec_factory = Recurrent
        inners = []
        for layer_idx in range(self.num_layers):
            idx = layer_idx * self.num_directions
            if self.bidirectional:
                layer_inners = [rec_factory(cells[idx]), rec_factory(cells[idx + 1], reverse=True)]
            else:
                layer_inners = [
                    rec_factory(cells[idx]),
                ]
            inners.extend(layer_inners)
        return StackedRNN(inners, self.num_layers, (self.mode == "LSTM"), dropout=self.dropout)

    def check_forward_args(self, input_, hidden, batch_sizes):
        is_input_packed = batch_sizes is not None
        expected_input_dim = 2 if is_input_packed else 3
        if input_.dim() != expected_input_dim:
            raise nncf.ValidationError(
                "input_ must have {} dimensions, got {}".format(expected_input_dim, input_.dim())
            )
        if self.input_size != input_.size(-1):
            raise nncf.ValidationError(
                "input_.size(-1) must be equal to input_size. Expected {}, got {}".format(
                    self.input_size, input_.size(-1)
                )
            )

        if is_input_packed:
            mini_batch = int(batch_sizes[0])
        else:
            mini_batch = input_.size(0) if self.batch_first else input_.size(1)

        expected_hidden_size = (mini_batch, self.hidden_size)

        def check_hidden_size(hx, expected_hidden_size, msg="Expected hidden size {}, got {}"):
            expected_size = self.num_layers * self.num_directions
            if expected_size != len(hx):
                raise nncf.InternalError("Expected number of hidden states {}, got {}".format(expected_size, len(hx)))
            for element in hx:
                if tuple(element.size()) != expected_hidden_size:
                    raise nncf.InternalError(msg.format(expected_hidden_size, tuple(element.size())))

        if self.mode == "LSTM":
            check_hidden_size(hidden[0], expected_hidden_size, "Expected hidden[0] size {}, got {}")
            check_hidden_size(hidden[1], expected_hidden_size, "Expected hidden[1] size {}, got {}")
        else:
            check_hidden_size(hidden, expected_hidden_size)

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.hidden_size)
        for weight in self.parameters():
            init.uniform_(weight, -stdv, stdv)

    @property
    def all_weights(self):
        return [[getattr(self, weight) for weight in weights] for weights in self._all_weights]

    @staticmethod
    def apply_permutation(tensor: torch.Tensor, permutation: torch.Tensor, dim: int = 1) -> torch.Tensor:
        return tensor.index_select(dim, permutation)

    def permute_hidden(
        self, hx: torch.Tensor, permutation: torch.Tensor
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        if permutation is None:
            return hx
        return self.apply_permutation(hx[0], permutation), self.apply_permutation(hx[1], permutation)

    def prepare_hidden(
        self, hx: torch.Tensor, permutation: torch.Tensor
    ) -> Union[torch.Tensor, Tuple[Tuple[torch.Tensor, ...], Tuple[torch.Tensor, ...]]]:
        if permutation is None:
            return hx
        split_size = len(hx[0])
        concat_hx = torch.cat([torch.unsqueeze(t, 0) for t in hx[0]])
        concat_cx = torch.cat([torch.unsqueeze(t, 0) for t in hx[1]])
        permuted_hidden = self.apply_permutation(concat_hx, permutation), self.apply_permutation(concat_cx, permutation)
        hc = permuted_hidden[0].chunk(split_size, 0)
        cc = permuted_hidden[1].chunk(split_size, 0)
        hidden = (tuple(torch.squeeze(c, 0) for c in hc), tuple(torch.squeeze(c, 0) for c in cc))
        return hidden

    def forward(self, input_, hidden=None):
        is_packed = isinstance(input_, PackedSequence)

        sorted_indices = None
        unsorted_indices = None
        if is_packed:
            input_, batch_sizes, sorted_indices, unsorted_indices = input_
            max_batch_size = int(batch_sizes[0])
        else:
            batch_sizes = None
            max_batch_size = input_.size(0) if self.batch_first else input_.size(1)

        if hidden is None:
            num_directions = 2 if self.bidirectional else 1
            hidden = torch.zeros(
                self.num_layers * num_directions,
                max_batch_size,
                self.hidden_size,
                requires_grad=False,
                device=input_.device,
            )
            if self.mode == "LSTM":
                hidden = (hidden, hidden)
        else:
            # Each batch of the hidden state should match the input sequence that
            # the user believes he/she is passing in.
            hidden = self.prepare_hidden(hidden, sorted_indices)

        self.check_forward_args(input_, hidden, batch_sizes)

        is_currently_variable = batch_sizes is not None
        if self.variable_length and not is_currently_variable or not self.variable_length and is_currently_variable:
            # override rnn_impl, it's assumed that this should happen very seldom, as
            # usually there's only one mode active whether variable length, or constant ones
            self.rnn_impl = self.get_rnn_impl(is_currently_variable, self.cells)

        if self.batch_first and batch_sizes is None:
            input_ = input_.transpose(0, 1)

        hidden, output = self.rnn_impl(input_, hidden, batch_sizes)

        if self.batch_first and batch_sizes is None:
            output = output.transpose(0, 1)

        if is_packed:
            output = PackedSequence(output, batch_sizes, sorted_indices, unsorted_indices)

        return output, self.permute_hidden(hidden, unsorted_indices)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import operator
from typing import Any, Optional, Tuple, TypeVar, Union

from nncf.experimental.tensor.definitions import TensorBackend
from nncf.experimental.tensor.definitions import TensorDataType
from nncf.experimental.tensor.definitions import TensorDeviceType

TTensor = TypeVar("TTensor")


class Tensor:
    """
    An interface to framework specific tensors for common NNCF algorithms.
    """

    def __init__(self, data: Optional[TTensor]):
        self._data = data.data if isinstance(data, Tensor) else data

    @property
    def data(self) -> TTensor:
        return self._data

    @property
    def shape(self) -> Tuple[int, ...]:
        return tuple(self.data.shape)

    @property
    def ndim(self) -> int:
        return self.data.ndim

    @property
    def device(self) -> TensorDeviceType:
        return _call_function("device", self)

    @property
    def dtype(self) -> TensorDataType:
        return _call_function("dtype", self)

    @property
    def backend(self) -> TensorBackend:
        return _call_function("backend", self)

    @property
    def size(self) -> int:
        return _call_function("size", self)

    def __bool__(self) -> bool:
        return bool(self.data)

    def __iter__(self):
        return TensorIterator(self.data)

    def __getitem__(self, index: Union[Tensor, int, Tuple[Union[Tensor, int], ...]]) -> Tensor:
        return Tensor(self.data[unwrap_index(index)])

    def __setitem__(self, index: Union[Tensor, int, Tuple[Union[Tensor, int], ...]], value: Any) -> None:
        self.data[unwrap_index(index)] = unwrap_tensor_data(value)

    def __str__(self) -> str:
        return f"nncf.Tensor({str(self.data)})"

    def __repr__(self) -> str:
        return f"nncf.Tensor({repr(self.data)})"

    # built-in operations

    def __add__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data + unwrap_tensor_data(other))

    def __radd__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(unwrap_tensor_data(other) + self.data)

    def __sub__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data - unwrap_tensor_data(other))

    def __rsub__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(unwrap_tensor_data(other) - self.data)

    def __mul__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data * unwrap_tensor_data(other))

    def __rmul__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(unwrap_tensor_data(other) * self.data)

    def __pow__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data ** unwrap_tensor_data(other))

    def __truediv__(self, other: Union[Tensor, float]) -> Tensor:
        return _call_function("_binary_op_nowarn", self, other, operator.truediv)

    def __rtruediv__(self, other: Union[Tensor, float]) -> Tensor:
        return _call_function("_binary_reverse_op_nowarn", self, other, operator.truediv)

    def __floordiv__(self, other: Union[Tensor, float]) -> Tensor:
        return _call_function("_binary_op_nowarn", self, other, operator.floordiv)

    def __rfloordiv__(self, other: Union[Tensor, float]) -> Tensor:
        return _call_function("_binary_reverse_op_nowarn", self, other, operator.floordiv)

    def __neg__(self) -> Tensor:
        return Tensor(-self.data)

    # Comparison operators

    def __lt__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data < unwrap_tensor_data(other))

    def __le__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data <= unwrap_tensor_data(other))

    def __eq__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data == unwrap_tensor_data(other))

    def __ne__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data != unwrap_tensor_data(other))

    def __gt__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data > unwrap_tensor_data(other))

    def __ge__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data >= unwrap_tensor_data(other))

    # Tensor functions

    def squeeze(self, axis: Optional[Union[int, Tuple[int, ...]]] = None) -> Tensor:
        return _call_function("squeeze", self, axis)

    def flatten(self) -> Tensor:
        return _call_function("flatten", self)

    def max(self, axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: Optional[bool] = False) -> Tensor:
        return _call_function("max", self, axis, keepdims)

    def min(self, axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: Optional[bool] = False) -> Tensor:
        return _call_function("min", self, axis, keepdims)

    def abs(self) -> Tensor:
        return _call_function("abs", self)

    def isempty(self) -> bool:
        return _call_function("isempty", self)

    def astype(self, dtype: TensorDataType) -> Tensor:
        return _call_function("astype", self, dtype)

    def reshape(self, shape: Tuple[int, ...]) -> Tensor:
        return _call_function("reshape", self, shape)

    def item(self) -> float:
        return _call_function("item", self)

    def clone(self) -> float:
        return _call_function("clone", self)


def _call_function(func_name: str, *args):
    """
    Call function from functions.py to avoid circular imports.

    :param func_name: Name of function.
    :return: Result of function call.
    """
    from nncf.experimental.tensor.functions import numeric

    fn = getattr(numeric, func_name)
    return fn(*args)


class TensorIterator:
    """Iterator for Tensor class"""

    def __init__(self, tensor):
        self._tensor = tensor
        self._index = 0

    def __next__(self) -> Tensor:
        if self._index < len(self._tensor):
            result = self._tensor[self._index]
            self._index += 1
            return Tensor(result)

        raise StopIteration


def unwrap_index(obj: Union[Any, Tuple[Any, ...]]) -> Union[TTensor, Tuple[TTensor, ...]]:
    """
    Unwraps the tensor data from the input object or tuple of objects.

    :param obj: The object to unwrap.
    :return: The unwrapped tensor data or tuple of unwrapped tensor data.
    """
    if isinstance(obj, tuple):
        return tuple(unwrap_tensor_data(o) for o in obj)
    return unwrap_tensor_data(obj)


def unwrap_tensor_data(obj: Any) -> TTensor:
    """
    Return the data of a Tensor object, or the object itself if it is not a Tensor.

    :param obj: The object to unwrap.
    :return: The data of the Tensor object, or the object itself.
    """
    return obj.data if isinstance(obj, Tensor) else obj



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from dataclasses import dataclass
from typing import Dict, List, Optional, Set


@dataclass
class ProducerInfo:
    """
    Defines a node that produce pruning masks. Operation that corresponds to the node has built-in weights.
    Pruning dimension is required in comparison with consumer info.

    :param node_id: id of the node that produce pruning
    :param pruning_dimension: axis number from 0 to N-1 in weights along which the pruning structure is defined.
        N is total number of dimensions.
    """

    node_id: int
    pruning_dimension: int = 0

    def __hash__(self) -> int:
        return hash((self.node_id, self.pruning_dimension))

    def __str__(self) -> str:
        return str(self.node_id)

    def __lt__(self, other: "ProducerInfo"):
        return self.node_id < other.node_id


@dataclass
class ConsumerInfo:
    """
    Defines the node that absorbs (consume) pruning masks. There are 2 types of consumer. First, have built-in weights,
    therefore can consume and produce pruning along the input channels. Consumers of the second type don't have built-in
    weights, instead weights are coming as a second input. In that case, pruning dimension is None.

    :param node_id: id of the node that consume pruning.
    :param pruning_dimension: axis number from 0 to N-1 in weights along which the pruning structure is defined.
        N is total number of dimensions. Equals None for the second type of consumers.
    """

    node_id: int
    pruning_dimension: Optional[int] = 1

    def __hash__(self) -> int:
        return hash((self.node_id, self.pruning_dimension))

    def __str__(self) -> str:
        return str(self.node_id)

    def __lt__(self, other: "ConsumerInfo"):
        return self.node_id < other.node_id


@dataclass
class PruningBlock:
    """
    Defines a pruning block - how much and which particular channels are supposed to be pruned at the same time for
    the node when a single element of pruning mask is 0. We assume that the pruning mask is a vector with 1's and 0's.
    1 retains the corresponding set of channels in weights, 0 prunes it.
    The block is initialized on producers of pruning and is propagated until consumers within PropagationGroup and
    PropagationMask.
    Essentially, the propagation block is parametrized by size and offset. Size is the number of sequent channels to be
    included to the block. When offset is not equal to 0, the block is formed by taking the `size` number of
    sequent channels, then skipping the `offset` number of channels and repeating the procedure for the rest channels.

    Let's consider 3 possible examples.
    Notation:
    "-" is not a pruned channel
    "x" is a channel that is included to the block or pruned

    1) By default, a single element of structured pruning mask corresponds to a single channel in the producer.
    Size is equal to 1, offset is 0. It means that block consist of just 1 channel.

    size = 1
    offset = 0
    number of all channels = 18

    all channels                        ------------------
    block #1, mask: 011111111111111111  x-----------------

    2) But in some cases, pruning can be performed by set of channels, rather than individual one.
    For instance, head pruning in the Transformers is removal of N sequent channels, when N is the size of the
    head. In that case, the propagation block is encoded by size=N and offset=0.

    size = 6 (head size)
    offset = 0
    number of all channels = 18

    all channels           ------------------
    block #1, mask: 011    xxxxxx------------
    block #2, mask: 101    ------xxxxxx------
    block #3, mask: 110    ------------xxxxxx
    mask: 100              ------xxxxxxxxxxxx

    3) Or potentially, one can remove the same dimension in each Transformer's head. The propagation block would be as
    follows: size=1 and offset=N. It means that block is formed by taking the `size` number of sequent channels, then
    skipping the `offset` number of sequent channels and repeating the procedure for the rest channels. For that
    particular case, if total number of channels is 3N, the size of block will be equal to 3.
    Block will contain 6 elements, if size=2 offset=N with the same size of pruning mask.

    size = 1
    offset = 6 (head size)
    number of all channels = 18

    all channels              ------------------
    block #1, mask: 011111    x-----x-----x-----
    block #2, mask: 101111    -x-----x-----x----
    ...
    block #6, mask: 111110    -----x-----x-----x
    mask 110000               --xxxx--xxxx--xxxx

    :param size: number of sequent channels.
    :param offset: when not equal to 0, block is formed by taking `size` number of sequent channels,
        then skipping `offset` number of sequent channels and repeating the procedure for the rest of channels.
    """

    size: int = 1
    offset: int = 0

    def __eq__(self, other) -> bool:
        return self.size == other.size and self.offset == other.offset

    def __str__(self) -> str:
        return f"S:{self.size}__O:{self.offset}"


class PropagationGroup:
    """
    Defines a group of propagation blocks and links it with the list of children groups.
    The group is initialized on producers of pruning and is propagated until consumers within PropagationMask.
    """

    def __init__(
        self,
        block: PruningBlock,
        producers: Optional[Set[ProducerInfo]] = None,
        consumers: Optional[Set[ConsumerInfo]] = None,
    ) -> None:
        self.block = block
        self._children: List["PropagationGroup"] = []
        self._is_invalid = False
        self._producers = set() if producers is None else producers
        self._consumers = set() if consumers is None else consumers

    @property
    def is_invalid(self):
        return self._is_invalid

    def __str__(self) -> str:
        producers = ",".join(map(str, sorted(self._producers)))
        consumers = ",".join(map(str, sorted(self._consumers)))
        return f"Block: {self.block}\nProducers: {producers}\nConsumers: {consumers}"

    def __repr__(self) -> str:
        producers = ",".join(map(str, self.get_producers()))
        consumers = ",".join(map(str, self.get_consumers()))
        return f"{str(self.block)}__P{producers}__C{consumers}"

    def invalidate(self) -> None:
        """
        Invalidate all blocks in the group and do the same for child groups.
        """
        self._is_invalid = True
        for child in self._children:
            child.invalidate()

    @staticmethod
    def join_groups(*args: "PropagationGroup") -> "PropagationGroup":
        """
        Join block groups into a new one. The group combines all block and child groups from the given list of groups.

        :return: a new block group.
        """
        first_block = args[0].block
        assert all(first_block == group.block for group in args), (
            "joining groups with different blocks is not "
            "supported. Need to implement merging of multiple block by choosing smallest common divisor as size.`"
        )

        new_group = PropagationGroup(block=first_block)
        for group in args:
            group.add_child(new_group)
            new_group.add_producers(group.get_producers())
            new_group.add_consumers(group.get_consumers())
        return new_group

    def add_consumers(self, consumers: Set[ConsumerInfo]):
        """
        Adds information about consumer to the group and to all children.
        The idea is to have up-to-date list of consumers in each leaf eventually.

        :param block: propagation block the corresponds to the consumer node.
        """
        self._consumers.update(consumers)
        for child in self._children:
            child.add_consumers(consumers)

    def add_producers(self, producers: Set[ProducerInfo]) -> None:
        self._producers.update(producers)

    def add_child(self, child: "PropagationGroup") -> None:
        self._children.append(child)

    def get_consumers(self) -> Set[ConsumerInfo]:
        return self._consumers.copy()

    def get_producers(self) -> Set[ProducerInfo]:
        return self._producers.copy()

    def get_children(self) -> List["PropagationGroup"]:
        return self._children.copy()

    def has_children(self) -> bool:
        return bool(self._children)


class PropagationMask:
    """
    Contains information about pruning in the current node - pruning block within propagation groups per dimension,
    for which they are applied.

    It's assumed that the propagation mask is initialized on producers and then propagated through the
    execution graph until consumer nodes.

    This helps to find possible ways of pruning nodes with tracking dependency between them.
    For example, to constrain a group of nodes to have the same structure or to find a
    specific pruning structure (PropagationBlock) that can be safely removed in all producers
    with retaining it in the consumers.
    """

    def __init__(self, dim_groups_map: Dict[int, List[PropagationGroup]] = None) -> None:
        self.dim_groups_map = dim_groups_map if dim_groups_map is not None else {}

    def __str__(self) -> str:
        state = {dim: list(map(repr, groups)) for dim, groups in self.dim_groups_map.items()}
        return json.dumps(state)

    def __bool__(self):
        return bool(self.dim_groups_map)

    def invalidate_groups(self) -> None:
        """
        Invalidate all blocks in the group and do the same for child groups.
        Can happen when propagation mask for some reason can't reach the consumer.
        For instance, when it's traversed until the node in ignored scope and that doesn't
        support pruning.
        """
        for groups in self.dim_groups_map.values():
            for group in groups:
                group.invalidate()



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List

import torch

from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.quantization.layers import BaseQuantizer


class QuantizerInfo:
    def __init__(self, quantizer_module_ref: BaseQuantizer, affected_insertions: List[PTTargetPoint]):
        self.quantizer_module_ref = quantizer_module_ref
        self.affected_insertions = affected_insertions


class NonWeightQuantizerInfo(QuantizerInfo):
    pass


class WeightQuantizerInfo(QuantizerInfo):
    def __init__(
        self,
        quantizer_module_ref: BaseQuantizer,
        quantized_module: torch.nn.Module,
        affected_insertions: List[PTTargetPoint],
    ):
        super().__init__(quantizer_module_ref, affected_insertions)
        self.quantized_module = quantized_module



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
import types
from collections import OrderedDict
from collections import defaultdict
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, TypeVar

import torch
from torch import nn

import nncf
from nncf import nncf_logger
from nncf.api.compression import CompressionAlgorithmController
from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.common.graph.definitions import MODEL_INPUT_OP_NAME
from nncf.common.graph.definitions import MODEL_OUTPUT_OP_NAME
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.operator_metatypes import CONST_NOOP_METATYPES
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.hook_handle import HookHandle
from nncf.common.insertion_point_graph import InsertionPointGraph
from nncf.common.insertion_point_graph import PostHookInsertionPoint
from nncf.common.insertion_point_graph import PreHookInsertionPoint
from nncf.common.utils.debug import is_debug
from nncf.torch.debug import CombinedDebugInterface
from nncf.torch.debug import debuggable_forward
from nncf.torch.dynamic_graph.context import PreHookId
from nncf.torch.dynamic_graph.context import TracingContext
from nncf.torch.dynamic_graph.graph import DynamicGraph
from nncf.torch.dynamic_graph.graph import ShapeIgnoringTensorMetaComparator
from nncf.torch.dynamic_graph.graph_tracer import GraphTracer
from nncf.torch.dynamic_graph.graph_tracer import WrapInputsFnType
from nncf.torch.dynamic_graph.graph_tracer import WrapOutputsFnType
from nncf.torch.dynamic_graph.graph_tracer import create_dummy_forward_fn
from nncf.torch.dynamic_graph.io_handling import InputInfoWrapManager
from nncf.torch.dynamic_graph.io_handling import ModelInputInfo
from nncf.torch.dynamic_graph.io_handling import replicate_same_tensors
from nncf.torch.dynamic_graph.io_handling import wrap_nncf_model_outputs_with_objwalk
from nncf.torch.dynamic_graph.operation_address import OperationAddress
from nncf.torch.dynamic_graph.patch_pytorch import ORIGINAL_CALL
from nncf.torch.dynamic_graph.scope import Scope
from nncf.torch.dynamic_graph.scope_access import get_module_by_scope
from nncf.torch.dynamic_graph.trace_functions import strip_traced_tensors
from nncf.torch.dynamic_graph.wrappers import wrap_module_call
from nncf.torch.dynamic_graph.wrappers import wrap_parameters
from nncf.torch.external_hook import EXTERNAL_OP_STORAGE_NAME
from nncf.torch.external_hook import ExternalOpCallHook
from nncf.torch.graph.graph import PTNNCFGraph
from nncf.torch.graph.graph_builder import GraphBuilder
from nncf.torch.graph.graph_builder import GraphConverter
from nncf.torch.graph.operator_metatypes import OPERATORS_WITH_WEIGHTS_METATYPES
from nncf.torch.graph.operator_metatypes import PTSplitMetatype
from nncf.torch.graph.transformations.command_creation import create_pt_insertion_command
from nncf.torch.graph.transformations.commands import DEFAULT_HOOKS_GROUP_NAME
from nncf.torch.graph.transformations.commands import ExtraCompressionModuleType
from nncf.torch.graph.transformations.commands import PTSharedFnInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.graph.transformations.serialization import serialize_transformations
from nncf.torch.knowledge_distillation.knowledge_distillation_handler import KnowledgeDistillationLossHandler
from nncf.torch.layer_utils import _NNCFModuleMixin
from nncf.torch.module_operations import UpdateWeight
from nncf.torch.nncf_module_replacement import replace_modules_by_nncf_modules
from nncf.torch.quantization.external_quantizer import EXTERNAL_QUANTIZERS_STORAGE_NAME
from nncf.torch.utils import compute_FLOPs_hook
from nncf.torch.utils import get_all_modules_by_type
from nncf.torch.utils import get_model_device
from nncf.torch.utils import training_mode_switcher

Module = TypeVar("Module", bound=nn.Module)


class PTInsertionType(IntEnum):
    NNCF_MODULE_PRE_OP = 0
    NNCF_MODULE_POST_OP = 1
    OPERATOR_PRE_HOOK = 2
    OPERATOR_POST_HOOK = 3


TARGET_TYPE_VS_PT_INSERTION_TYPE_DICT_FOR_REPLACED_MODULES = {
    TargetType.PRE_LAYER_OPERATION: PTInsertionType.NNCF_MODULE_PRE_OP,
    TargetType.POST_LAYER_OPERATION: PTInsertionType.NNCF_MODULE_POST_OP,
    TargetType.OPERATION_WITH_WEIGHTS: PTInsertionType.NNCF_MODULE_PRE_OP,
    TargetType.OPERATOR_PRE_HOOK: PTInsertionType.OPERATOR_PRE_HOOK,
    TargetType.OPERATOR_POST_HOOK: PTInsertionType.OPERATOR_POST_HOOK,
}

TARGET_TYPE_VS_PT_INSERTION_TYPE_DICT_FOR_NOT_REPLACED_MODULES = {
    TargetType.PRE_LAYER_OPERATION: PTInsertionType.OPERATOR_PRE_HOOK,
    TargetType.POST_LAYER_OPERATION: PTInsertionType.OPERATOR_POST_HOOK,
    TargetType.OPERATION_WITH_WEIGHTS: PTInsertionType.OPERATOR_PRE_HOOK,
    TargetType.OPERATOR_PRE_HOOK: PTInsertionType.OPERATOR_PRE_HOOK,
    TargetType.OPERATOR_POST_HOOK: PTInsertionType.OPERATOR_POST_HOOK,
}


class PTInsertionPoint:
    def __init__(
        self,
        target_type: TargetType,
        op_address: OperationAddress,
        input_port_id: int = None,
        replaced_modules: bool = True,
    ):
        self.insertion_type = self._get_pt_insertion_type(target_type, replaced_modules)
        self.op_address = op_address
        self.module_scope = op_address.scope_in_model
        self.input_port_id = input_port_id

    @staticmethod
    def _get_pt_insertion_type(target_type: TargetType, replaced_modules: bool) -> PTInsertionType:
        map_target_types = TARGET_TYPE_VS_PT_INSERTION_TYPE_DICT_FOR_NOT_REPLACED_MODULES
        if replaced_modules:
            map_target_types = TARGET_TYPE_VS_PT_INSERTION_TYPE_DICT_FOR_REPLACED_MODULES

        if not isinstance(target_type, TargetType) or target_type not in map_target_types:
            raise nncf.InternalError("Unsupported target type for PyTorch: {}".format(target_type))
        return map_target_types[target_type]

    def __eq__(self, other: "PTInsertionPoint"):
        return (
            self.insertion_type == other.insertion_type
            and self.op_address == other.op_address
            and self.module_scope == other.module_scope
            and self.input_port_id == other.input_port_id
        )

    def __str__(self):
        return " ".join([str(v) for v in self.__dict__.values()])

    def __hash__(self):
        return hash(str(self))


@dataclass
class PTGraphPair:
    """
    Container for two dependent graph representation:
    DynamicGraph and NNCFGraph built out of DynamicGraph.
    """

    dynamic_graph: DynamicGraph
    nncf_graph: NNCFGraph


class NNCFNetworkInterface(torch.nn.Module):
    """
    The single object that is added to the original model object as an attribute to provide a namespace for
    NNCF-specific method calls and a torch.nn.Module-like storage for compression parameters. Since this is a
    Module stored in a Module, all trainable parameters of the NNCFInterface will be registered for optimization
    in the same manner as the original model parameters, and will also be eligible for state_dict-powered persistence
    when saving/loading checkpoints
    """

    MODEL_STATE_VERSION_ATTR = "_nncf_model_state_version"
    MODEL_STATE_VERSION = 1

    def forward(self):
        """
        The module only serves a storage and namespacing purpose, forward functionality is not implemented.
        """
        raise NotImplementedError("Calling `forward` on NNCFInterface is prohibited.")

    def get_original_forward(self) -> Callable:
        """
        Returns the forward function of the original model, unmodified by NNCF. The returned function will
        have its 0-th implicit `self` argument bound to the model object.
        """
        if self._original_instance_forward is not None:
            return self._original_instance_forward
        return functools.partial(self._original_unbound_forward, self._model_ref)

    @contextmanager
    def temporary_bound_original_forward(self, bound_forward: Callable):
        """
        Context manager for temporary replacement of the underlying original model forward function. NNCF
        works by doing additional operations before and after the original object's forward call, and this context
        manager allows to temporarily run the compressed model object as if it had another original forward method.
        The signature of the new forward method must be the same w.r.t. the original forward method in terms of
        activation tensors.
        :param bound_forward: A callable which will be used to temporary replace the original forward call. Must be
        a bound method, e.g. the `self` argument had already been set to the same model object where the forward call
        must be replaced.
        """
        prev_bound_forward = self._bound_original_forward
        self._bound_original_forward = bound_forward
        yield
        self._bound_original_forward = prev_bound_forward

    def get_original_unbound_forward(self) -> Callable:
        """
        Returns the forward function of the original model, unmodified by NNCF. The returned function will preserve
        its 0-th implicit `self` argument without binding it to the model object.
        """
        return self._original_unbound_forward

    def set_original_unbound_forward(self, fwd_fn: Callable):
        """
        Allows to set the function that is treated by NNCF as "original" model forward to another function.
        :param fwd_fn: The new original forward function. The signature w.r.t. activation tensors must be the same,
        and the function must leave its 0-th `self` argument unbound.
        """
        self._custom_original_unbound_forward = fwd_fn

    def reset_original_unbound_forward(self):
        """
        Reset the forward which was set with set_original_unbound_forward() method.
        After this NNCF will fall back to the unbound forward of the original model.
        """
        self._custom_original_unbound_forward = None

    def __init__(
        self,
        model: torch.nn.Module,
        input_info: ModelInputInfo = None,
        dummy_forward_fn: Callable = None,
        wrap_inputs_fn: WrapInputsFnType = None,
        scopes_without_shape_matching: List[str] = None,
        ignored_scopes: List[str] = None,
        target_scopes: List[str] = None,
        wrap_outputs_fn: WrapOutputsFnType = None,
        replace_modules: bool = True,
        trace_parameters: bool = False,
    ):
        super().__init__()

        self.replace_modules = replace_modules
        self.trace_parameters = trace_parameters

        # Need this in order not to register owning module as sub-module of NNCFInterface and thus
        # avoid circular references
        object.__setattr__(self, "__model_ref", model)

        if isinstance(model, NNCFNetwork):
            # Got an NNCFNetwork already, probably during shallow copying.
            self._original_class = model.nncf._original_class
            self._bound_original_forward = model.nncf._bound_original_forward
            self._custom_original_unbound_forward = model.nncf._custom_original_unbound_forward
            self._original_instance_forward = model.nncf._original_instance_forward
        else:
            self._original_class = model.__class__
            self._bound_original_forward = None
            self._custom_original_unbound_forward = None
            self._original_instance_forward = model.__dict__.get("forward")

        self._forward_signature = inspect.signature(self.get_original_forward())
        self._input_info = input_info

        self._ignored_scopes = ignored_scopes
        self._target_scopes = target_scopes
        self._user_dummy_forward_fn = dummy_forward_fn
        self._kd_loss_handler = None
        self._groups_vs_hooks_handlers: Dict[str, List[HookHandle]] = defaultdict(list)

        if wrap_inputs_fn is not None:
            self._wrap_inputs_fn = wrap_inputs_fn
        elif self._input_info is not None:
            self.__input_info_based_input_wrapper = InputInfoWrapManager(
                self._input_info, self._forward_signature, module_ref_for_device=model
            )
            self._wrap_inputs_fn = self.__input_info_based_input_wrapper.wrap_inputs
        else:
            raise ValueError("wrap_inputs_fn or input_infos should be passed.")

        if wrap_outputs_fn is not None:
            self._wrap_outputs_fn = wrap_outputs_fn
        else:
            self._wrap_outputs_fn = wrap_nncf_model_outputs_with_objwalk

        self._nncf_replaced_modules: Dict[torch.nn.Module, List[Scope]] = {}
        self._scopes_without_shape_matching = scopes_without_shape_matching
        self.debug_interface = CombinedDebugInterface() if is_debug() else None
        self._extra_module_types: List[ExtraCompressionModuleType] = []
        self._insertions_into_original_graph: Dict[PTTargetPoint, List[Tuple[Callable, TransformationPriority]]] = {}

        if isinstance(model, NNCFNetwork):
            self._nncf_replaced_modules = model.nncf._nncf_replaced_modules
        else:
            _orig_graph_build_forward_fn = self._get_dummy_forward_fn_for_graph_building(
                with_input_tracing=True, with_output_tracing=True
            )

            if self.replace_modules:
                eval_op_scopes = self._collect_eval_op_scopes(model, _orig_graph_build_forward_fn)

                # all modules called in eval mode should be replaced prior to graph building
                self._replace_modules_by_nncf_modules(model, eval_op_scopes)

        _orig_context = TracingContext()

        _orig_context.add_node_comparators([MODEL_INPUT_OP_NAME], ShapeIgnoringTensorMetaComparator())
        _orig_context.add_node_comparators([MODEL_OUTPUT_OP_NAME], ShapeIgnoringTensorMetaComparator())
        if self._scopes_without_shape_matching:
            _orig_context.add_node_comparators(scopes_without_shape_matching, ShapeIgnoringTensorMetaComparator())

        if isinstance(model, NNCFNetwork):
            self._original_graphs_pair = model.nncf._original_graphs_pair
        else:
            original_dynamic_graph = GraphTracer(_orig_graph_build_forward_fn).trace_graph(
                model, _orig_context, as_eval=True, trace_parameters=self.trace_parameters
            )
            original_graph = GraphConverter.convert(original_dynamic_graph, trace_parameters)
            self._original_graphs_pair = PTGraphPair(dynamic_graph=original_dynamic_graph, nncf_graph=original_graph)
        self._compressed_graphs_pair: PTGraphPair = None

        self._compressed_context = TracingContext()

        if self.trace_parameters:
            self._compressed_context.reused_parameters = self.get_reused_parameters()

        self._dummy_forward_fn = self._get_dummy_forward_fn_for_graph_building(
            with_input_tracing=False, with_output_tracing=False
        )
        self._in_user_dummy_forward = False

        self._compressed_context.add_node_comparators([MODEL_INPUT_OP_NAME], ShapeIgnoringTensorMetaComparator())
        self._compressed_context.add_node_comparators([MODEL_OUTPUT_OP_NAME], ShapeIgnoringTensorMetaComparator())
        if self._scopes_without_shape_matching:
            self._compressed_context.add_node_comparators(
                scopes_without_shape_matching, ShapeIgnoringTensorMetaComparator()
            )
        self._load_listener = None

        self.compression_controller: CompressionAlgorithmController = None

    @property
    def _original_unbound_forward(self):
        # Notes:
        # (1) We rely on an "unbound" forward which is the version of the method that has the
        #   `self` parameter not set, otherwise we will be indirectly capturing a reference to the
        #   model object in NNCFInterface - this will lead to failures in DataParallel
        #   because the bound original forward call during NNCFNetwork.forward
        #   would then call forward on the original non-replica module even if NNCFNetwork itself was
        #   replicated.
        # (2) We access the unbound forward from a reference to the original model class instead
        #   of storing the reference to the unbound forward itself because the original class forward
        #   may be overridden by some 3rd party logic. For example, during export of mm-based models to ONNX
        #   using mmdeploy library, the original forward method of the model is temporarily replaced
        #   during export. Moreover, in such case the forward signature needs to be hidden by a user
        #   beforehand by wrapping it with a function with (*args, **kwargs) as its arguments.
        custom_unbound_forward = self._custom_original_unbound_forward
        return self._original_class.forward if custom_unbound_forward is None else custom_unbound_forward

    @property
    def _model_ref(self) -> "NNCFNetwork":
        return object.__getattribute__(self, "__model_ref")

    @property
    def input_infos(self) -> ModelInputInfo:
        return deepcopy(self._input_info)

    def create_knowledge_distillation_loss_handler(
        self, kd_original_model: nn.Module, calculate_fn
    ) -> KnowledgeDistillationLossHandler:
        """
        Creates KnowledgeDistillationLossHandler instance for enabling Knowledge Distillation feature.
            Also returns created KnowledgeDistillationLossHandler for control over Knowledge Distillation logic.

        :param kd_original_model: original non compressed model used for distillation
        :param calculate_fn: function used to parse model outputs and calculate knowledge distillation loss
        :return: KnowledgeDistillationLossHandler instance
        """
        device = get_model_device(self._model_ref)
        self._kd_loss_handler = KnowledgeDistillationLossHandler(
            self._compressed_context, kd_original_model, calculate_fn, device
        )
        return self._kd_loss_handler

    def reset_nncf_modules(self):
        for scope_list in self.get_nncf_module_scopes():
            # Can pick any access scope since they all should
            # point to the same object
            some_scope = scope_list[0]
            module = self.get_module_by_scope(some_scope)
            module.reset()

    def get_clean_shallow_copy(self) -> "NNCFNetwork":
        # WARNING: Will reset pre- and post-ops of the underlying model. Use save_nncf_module_additions
        # and load_nncf_module_additions to preserve these, or temporary_clean_view().
        from nncf.torch.utils import load_module_state
        from nncf.torch.utils import save_module_state

        saved_state = save_module_state(self._model_ref)
        new_interface = NNCFNetworkInterface(
            self._model_ref,
            self._input_info,
            self._user_dummy_forward_fn,
            self._wrap_inputs_fn,
            self._scopes_without_shape_matching,
            self._ignored_scopes,
            self._target_scopes,
            wrap_outputs_fn=self._wrap_outputs_fn,
            replace_modules=self.replace_modules,
            trace_parameters=self.trace_parameters,
        )
        self._model_ref._nncf = new_interface
        self._model_ref.nncf.reset_nncf_modules()
        load_module_state(self._model_ref, saved_state)
        return self._model_ref

    def get_modules_in_nncf_modules_by_type(self, class_names: List[str]) -> Dict[Scope, nn.Module]:
        nncf_modules = self.get_nncf_modules()
        retval = {}
        for nncf_module, nncf_module_scope in nncf_modules.items():
            nncf_module_scope.pop()
            for relative_scope, target_module in get_all_modules_by_type(nncf_module, class_names).items():
                retval[nncf_module_scope + relative_scope] = target_module
        return retval

    def insert_at_point(
        self,
        point: PTInsertionPoint,
        fn: Callable,
        hooks_group_name: Optional[str] = DEFAULT_HOOKS_GROUP_NAME,
    ) -> List[HookHandle]:
        """
        Inserts given function to the point in the NNCFNetwork, creates hook handle for the inserted function and
        stores created hook handle in a group with the given name. A group name could be used late
        to remove all hooks from the NNCFNetwork which belongs to the group.

        :param point: Target point to insert function.
        :param fn: Function to insert to the NNCFNetwork.
        :param hooks_group_name: Name of hooks group for hook handle associated with the inserted function.
        :return: Hook handle associated with the inserted function.
        """
        handle = None
        if point.insertion_type == PTInsertionType.OPERATOR_PRE_HOOK:
            handle = self._compressed_context.register_pre_hook(fn, point.op_address, point.input_port_id)
        elif point.insertion_type == PTInsertionType.OPERATOR_POST_HOOK:
            handle = self._compressed_context.register_post_hook(fn, point.op_address)
        elif point.insertion_type in [PTInsertionType.NNCF_MODULE_PRE_OP, PTInsertionType.NNCF_MODULE_POST_OP]:
            nncf_module = self.get_module_by_scope(point.module_scope)
            if not isinstance(nncf_module, _NNCFModuleMixin):
                raise nncf.ValidationError(
                    f"Failed to insert pre/post op for not registered custom module {point.module_scope}. NNCF only "
                    f"supports native PyTorch modules with respect to trainable parameter (weight) compressed, such "
                    f"as `torch.nn.Conv2d`. If your model contains a custom, non-PyTorch standard module with trainable"
                    f" weights that should be compressed, you can register it using the "
                    f"`@nncf.register_module` decorator. Please refer to `Compression of custom modules` section in "
                    f"docs/Usage.md for more details."
                )

            norm_target_scope = self._normalize_variable_recurrent_scope(point.module_scope)
            norm_nncf_scopes = []
            for scope_list_for_module in self.get_nncf_module_scopes():
                norm_nncf_scopes.extend([self._normalize_variable_recurrent_scope(x) for x in scope_list_for_module])
            assert norm_target_scope in norm_nncf_scopes  # Required for proper Recurrent/VariableRecurrent addressing
            if point.insertion_type == PTInsertionType.NNCF_MODULE_PRE_OP:
                handle = nncf_module.register_pre_forward_operation(fn)
            elif point.insertion_type == PTInsertionType.NNCF_MODULE_POST_OP:
                handle = nncf_module.register_post_forward_operation(fn)
        else:
            raise nncf.ValidationError("Unsupported insertion type: {}".format(point.insertion_type))
        self._groups_vs_hooks_handlers[hooks_group_name].append(handle)
        return handle

    def remove_hooks_group(self, hooks_group_name: str) -> None:
        """
        Removes all hooks of given hooks group from the nncf interface.

        :param group: Target hooks group name to remove all hooks from.
        """
        for handle in self._groups_vs_hooks_handlers[hooks_group_name]:
            handle.remove()
        del self._groups_vs_hooks_handlers[hooks_group_name]

    def get_graph(self) -> PTNNCFGraph:
        if self._compressed_context.graph.get_nodes_count() == 0 or self._compressed_graphs_pair.nncf_graph is None:
            self.rebuild_graph()
        return self._compressed_graphs_pair.nncf_graph

    def get_dynamic_graph(self) -> DynamicGraph:
        return self._compressed_context.graph

    def get_original_graph(self) -> PTNNCFGraph:
        return self._original_graphs_pair.nncf_graph

    def get_tracing_context(self) -> TracingContext:
        return self._compressed_context

    def enable_dynamic_graph_building(self):
        self._compressed_context.enable_node_additions()

    def disable_dynamic_graph_building(self):
        self._compressed_context.disable_node_additions()

    def _get_dummy_forward_fn_for_graph_building(self, with_input_tracing, with_output_tracing):
        if self._user_dummy_forward_fn is None:
            return create_dummy_forward_fn(
                self._input_info,
                with_input_tracing=with_input_tracing,
                wrap_inputs_fn=self._wrap_inputs_fn,
                wrap_outputs_fn=self._wrap_outputs_fn,
                with_output_tracing=with_output_tracing,
            )

        def wrapped_user_dummy_forward_fn(*args, **kwargs):
            self._in_user_dummy_forward = True
            retval = self._user_dummy_forward_fn(*args, **kwargs)
            self._in_user_dummy_forward = False
            return retval

        return wrapped_user_dummy_forward_fn

    def _replace_modules_by_nncf_modules(self, model: torch.nn.Module, eval_op_scopes: List[Scope] = None):
        _, self._nncf_replaced_modules = replace_modules_by_nncf_modules(
            model, ignored_scopes=self._ignored_scopes, target_scopes=self._target_scopes, eval_op_scopes=eval_op_scopes
        )
        return model

    def get_nncf_module_scopes(self) -> List[List[Scope]]:
        return list(self._nncf_replaced_modules.values())

    def get_nncf_modules(self) -> Dict[torch.nn.Module, Scope]:
        retval = {}
        for module, scope_set in self._nncf_replaced_modules.items():
            canonical_scope = next(iter(scope_set))
            retval[module] = canonical_scope.copy()
        return retval

    def get_weighted_original_graph_nodes(self, nncf_module_names: List[str] = None) -> List[NNCFNode]:
        retval = set()
        for scope_list in self.get_nncf_module_scopes():
            for nncf_module_scope in scope_list:
                if nncf_module_names is not None:
                    module_name = nncf_module_scope[-1].calling_module_class_name
                    if module_name not in nncf_module_names:
                        continue
                nncf_graph: PTNNCFGraph = self._original_graphs_pair.nncf_graph
                nodes_in_scope = nncf_graph.get_op_nodes_in_scope(nncf_module_scope)
                for node in nodes_in_scope:
                    if node.metatype in OPERATORS_WITH_WEIGHTS_METATYPES:
                        retval.add(node)

        return list(sorted(retval, key=str))

    def rebuild_graph(self, *input_args):
        self._compressed_context.reset_graph()
        dummy_forward_fn = self._get_dummy_forward_fn_for_graph_building(
            with_input_tracing=False, with_output_tracing=False
        )
        builder = GraphBuilder(dummy_forward_fn)

        with training_mode_switcher(self._model_ref, is_training=False):
            compressed_traced_graph = builder.build_dynamic_graph(
                self._model_ref, self._compressed_context, trace_parameters=self.trace_parameters
            )
            compressed_graph = GraphConverter.convert(compressed_traced_graph, self.trace_parameters)
            self._compressed_graphs_pair = PTGraphPair(
                dynamic_graph=compressed_traced_graph, nncf_graph=compressed_graph
            )

    def is_scope_in_nncf_module_scope(self, scope: Scope) -> bool:
        norm_nncf_scopes = []
        for scope_list_for_module in self.get_nncf_module_scopes():
            norm_nncf_scopes.extend([self._normalize_variable_recurrent_scope(x) for x in scope_list_for_module])
        norm_op_scope = self._normalize_variable_recurrent_scope(scope)
        for nncf_scope in norm_nncf_scopes:
            if norm_op_scope in nncf_scope:
                return True
        return False

    def register_compression_module_type(self, compression_module_type: ExtraCompressionModuleType):
        attr_name = compression_module_type_to_attr_name(compression_module_type)
        if compression_module_type in self._extra_module_types:
            raise nncf.ValidationError(f"Module type {compression_module_type} is already registered")

        self.__setattr__(attr_name, nn.ModuleDict())
        self._extra_module_types.append(compression_module_type)

    def add_compression_module(
        self, module_key: str, module: nn.Module, compression_module_type: ExtraCompressionModuleType
    ):
        attr_name = compression_module_type_to_attr_name(compression_module_type)
        if compression_module_type not in self._extra_module_types:
            raise nncf.InternalError(f"Module type {compression_module_type} was not registered")
        storage = self.__getattr__(attr_name)
        if module_key in storage:
            raise nncf.InternalError(f"Module {module_key} is already registered under {attr_name}")
        storage[module_key] = module

    def get_compression_modules_by_type(self, compression_module_type: ExtraCompressionModuleType) -> nn.ModuleDict:
        attr_name = compression_module_type_to_attr_name(compression_module_type)
        if compression_module_type not in self._extra_module_types:
            raise nncf.InternalError(f"Module type {compression_module_type} was not registered")
        return self.__getattr__(attr_name)

    def is_compression_module_registered(self, compression_module_type: ExtraCompressionModuleType) -> bool:
        """
        Check that extra compression module was registered.

        :param compression_module_type: Type of the extra compression module.
        :return: True if the extra compression module is registered, otherwise False.
        """
        return compression_module_type in self._extra_module_types

    def sort_compression_modules(self, compression_module_type: ExtraCompressionModuleType):
        attr_name = compression_module_type_to_attr_name(compression_module_type)
        if compression_module_type not in self._extra_module_types:
            raise nncf.InternalError("Module type {} was not registered".format(compression_module_type))
        module_dict = self.__getattr__(attr_name)

        module_dict._modules = OrderedDict(sorted(module_dict._modules.items()))
        self.__setattr__(attr_name, module_dict)

    @staticmethod
    def _normalize_variable_recurrent_scope(scope: Scope):
        """
        Two scopes pointing to an NNCF module that only differ in a Recurrent/VariableRecurrent/VariableRecurrentReverse
        scope node actually point to one and the same module.
        """
        ret_scope = scope.copy()
        for scope_element in ret_scope:
            if scope_element.calling_module_class_name in [
                "Recurrent",
                "VariableRecurrent",
                "VariableRecurrentReverse",
            ]:
                scope_element.calling_module_class_name = "NormalizedName_Recurrent"
        return ret_scope

    def do_dummy_forward(self, force_eval: bool = False):
        """
        Attention: If run with force_eval=False, this may spoil the batchnorm statistics,
        and an eval run of the model will perform much worse than the train run.
        """
        if force_eval:
            train_mode = self._model_ref.training
            self._model_ref.eval()
        with torch.no_grad():
            with self._compressed_context as ctx:
                ctx.base_module_thread_local_replica = self._model_ref
                self._dummy_forward_fn(self._model_ref)
        if force_eval and train_mode:
            self._model_ref.train()

    def get_original_insertion_point_graph(self) -> InsertionPointGraph:
        # Set up a pre- and post-hooks on almost every op in PyTorch
        nncf_graph = self.get_original_graph()
        pre_hooks: List[PreHookInsertionPoint] = []
        post_hooks: List[PostHookInsertionPoint] = []
        for node in nncf_graph.get_all_nodes():
            # Pre-hook insertion point nodes
            # Will insert a pre-hook IP for each input edge. The input edge must be marked with
            # a port ID attribute.
            in_edges = nncf_graph.get_input_edges(node)
            for edge in in_edges:
                for port_id in [
                    edge.input_port_id,
                ] + edge.parallel_input_port_ids:
                    pre_hook_ip = PreHookInsertionPoint(target_node_name=node.node_name, input_port_id=port_id)
                    pre_hooks.append(pre_hook_ip)

            if issubclass(node.metatype, PTSplitMetatype):
                # chunk returns a tuple of tensors, which can only be handled in NNCF
                # once post-hook ports are enabled. Work around it for now by disallowing post-hook
                # insertion for chunks
                # TODO: enable post-hook ports and remove this
                continue

            # Post-hook insertion point nodes
            post_hook_ip = PostHookInsertionPoint(node.node_name)
            post_hooks.append(post_hook_ip)

        weighted_nodes = self.get_weighted_original_graph_nodes()
        weighted_node_names = [weighted_node.node_name for weighted_node in weighted_nodes]

        ip_graph = InsertionPointGraph(
            self._original_graphs_pair.nncf_graph,
            weight_modifiable_node_names=weighted_node_names,
            allowed_pre_hook_insertion_points=pre_hooks,
            allowed_post_hook_insertion_points=post_hooks,
        )
        return ip_graph

    def get_module_by_scope(self, scope: Scope) -> Optional[torch.nn.Module]:
        curr_module = self._model_ref
        return get_module_by_scope(curr_module, scope)

    def get_containing_module(self, node_name: NNCFNodeName) -> torch.nn.Module:
        if self._compressed_graphs_pair is not None:
            try:
                scope = self._compressed_graphs_pair.nncf_graph.get_scope_by_node_name(node_name)
            except RuntimeError:
                nncf_logger.debug(
                    f"Node {node_name} not found in compressed graph when trying to determine "
                    f"the containing module, trying the original graph to see if the node was "
                    f"present there during graph building"
                )
                scope = self._original_graphs_pair.nncf_graph.get_scope_by_node_name(node_name)
        else:
            scope = self._original_graphs_pair.nncf_graph.get_scope_by_node_name(node_name)
        return self.get_module_by_scope(scope)

    def get_flops_per_module(self) -> Dict[NNCFNodeName, int]:
        """
        Calculates FLOPS count for modules.
        """
        model = self._model_ref
        flops_count_dict = {}

        def get_hook(name):
            return functools.partial(compute_FLOPs_hook, dict_to_save=flops_count_dict, module_node_name=name)

        hook_list = []
        for nncf_node in self._original_graphs_pair.nncf_graph.get_all_nodes():
            node_module = self.get_containing_module(nncf_node.node_name)
            hook_list.append(node_module.register_forward_hook(get_hook(nncf_node.node_name)))
        model.nncf.do_dummy_forward(force_eval=True)

        for h in hook_list:
            h.remove()
        return flops_count_dict

    def get_MACs_in_model(self):
        """
        Calculates MAC units count for model.
        """
        flops_count_dict = self.nncf.get_flops_per_module()
        total_MACs_count = sum(v // 2 for v in flops_count_dict.values())
        return total_MACs_count

    def save_nncf_module_additions(self) -> Dict[Scope, Tuple[torch.nn.ModuleDict, torch.nn.ModuleDict]]:
        retval = {}
        for nncf_module, module_scope in self.get_nncf_modules().items():
            retval[module_scope] = (deepcopy(nncf_module.pre_ops), deepcopy(nncf_module.post_ops))
        return retval

    def load_nncf_module_additions(
        self, scope_vs_pre_post_ops_dict: Dict[Scope, Tuple[torch.nn.ModuleDict, torch.nn.ModuleDict]]
    ):
        for nncf_module, module_scope in self.get_nncf_modules().items():
            nncf_module.pre_ops = scope_vs_pre_post_ops_dict[module_scope][0]
            nncf_module.post_ops = scope_vs_pre_post_ops_dict[module_scope][1]

    def temporary_clean_view(self):
        class Mgr:
            def __init__(self, model: NNCFNetwork):
                self.model = model
                self.nncf_module_state_dicts = {}
                self.nncf_interface_state_dict = None
                self.nncf_compression_module_types = []

            def __enter__(self):
                self.nncf_module_state_dicts = self.model.nncf.save_nncf_module_additions()
                self.nncf_interface = self.model.nncf
                clean_model = self.model.nncf.get_clean_shallow_copy()
                return clean_model

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.model._nncf = self.nncf_interface
                self.model.nncf.load_nncf_module_additions(self.nncf_module_state_dicts)

        return Mgr(self._model_ref)

    def _collect_eval_op_scopes(self, model: nn.Module, dummy_forward_fn: Callable) -> List[Scope]:
        """
        Returns scopes of the operations in the graph which are executed in evaluation mode.
        """

        tracer = GraphTracer(dummy_forward_fn)
        result = []
        eval_graph = tracer.trace_graph(model, as_eval=True)
        root_scope = Scope()
        for dyn_graph_node in eval_graph.get_all_nodes():
            scope_in_model = dyn_graph_node.op_exec_context.scope_in_model
            if scope_in_model != root_scope:  # happens for ops such as /nncf_model_input_* and /nncf_model_output_*
                result.append(scope_in_model)
        return result

    def get_config(self) -> Dict[str, Any]:
        """
        Returns serializable NNCFNetwork config which contains
        all information required to recover all additional modules placement.

        :return: Serializable config which contains
            all information required to recover all additional modules placement.
        """
        transformation_layout = self.transformation_layout()
        config = serialize_transformations(transformation_layout)
        config[NNCFNetwork.TRACE_PARAMETERS_KEY] = self.trace_parameters
        return config

    def transformation_layout(self) -> PTTransformationLayout:
        """
        Collects all hooks applied to the NNCFNetwork, converts them to insertion commands
        and returns in PTTransformationLayout format. Default hooks group name is used in
        recovered commands, so hooks group names specified during the model modification
        become outdated.

        :return: Transformation layout with all commands applied to the NNCFNetwork.
        """

        def _check_external_call_hook_is_valid(hook: ExternalOpCallHook, info: str):
            """
            Check given external op call hook reference is correct.

            :param hook: External op call hook to check correctness.
            :param info: Info to log in case op call hook references are broken.
            """
            assert hasattr(
                self, hook._storage_name
            ), f"Storage name {hook._storage_name} is not registered. Info: {info}"
            assert hook._storage_key in getattr(
                self, hook._storage_name
            ), f"Key {hook._storage_key} is not registered in {hook._storage_name}. Info: {info}"

        context_hooks = defaultdict(lambda: defaultdict(list))
        transformation_layout = PTTransformationLayout()
        nncf_graph = self.get_graph()
        nncf_node_names_map = self.get_op_address_to_op_name_map()

        # Collect pre/post layer and op with weights insertion commands
        for nncf_module, module_scope in self.get_nncf_modules().items():
            for ops, target_type in (
                (nncf_module.pre_ops, TargetType.PRE_LAYER_OPERATION),
                (nncf_module.post_ops, TargetType.POST_LAYER_OPERATION),
            ):
                for priority, module in enumerate(ops.values()):
                    nodes_in_scope = nncf_graph.get_op_nodes_with_scope(module_scope)
                    # Several NNCFNodes means that current NNCFModule was called
                    # several times. Only one insertion command is required to
                    # call hook as much times as the current NNCFModule, therefore
                    # we use first correspondent NNCFNode.
                    nncf_node = nodes_in_scope[0]
                    command_target_type = target_type
                    if isinstance(module, UpdateWeight):
                        command_target_type = TargetType.OPERATION_WITH_WEIGHTS
                        module = module.op
                    if not isinstance(module, ExternalOpCallHook):
                        command = create_pt_insertion_command(
                            module, command_target_type, nncf_node.node_name, priority, None
                        )
                        transformation_layout.register(command)
                        continue

                    info = (
                        f"TargetType: {command_target_type}, nncf node name: {nncf_node.node_name},"
                        f" priority: {priority}, fn: {module}"
                    )
                    _check_external_call_hook_is_valid(module, info)

                    context_hooks[module._storage_name][module._storage_key].append(
                        (command_target_type, nncf_node.node_name, priority, module, None)
                    )

        # Collect all pre/post hooks commands
        for ops, target_type in (
            (self._compressed_context._pre_hooks, TargetType.OPERATOR_PRE_HOOK),
            (self._compressed_context._post_hooks, TargetType.OPERATOR_POST_HOOK),
        ):
            for op_address, hooks in ops.items():
                if isinstance(op_address, PreHookId):
                    input_port_id = op_address.input_port_id
                    op_address = op_address.op_address
                else:
                    input_port_id = None
                for priority, fn in enumerate(hooks.values()):
                    target_node_names = nncf_node_names_map[op_address]
                    # Operation address is unique for each module call
                    assert len(target_node_names) == 1
                    target_node_name = target_node_names[0]

                    if not isinstance(fn, ExternalOpCallHook):
                        command = create_pt_insertion_command(
                            fn, target_type, target_node_name, priority, input_port_id
                        )
                        transformation_layout.register(command)
                        continue

                    info = f"TargetType: {target_type}, op_address: {op_address}, priority: {priority}, fn: {fn}"
                    _check_external_call_hook_is_valid(fn, info)

                    context_hooks[fn._storage_name][fn._storage_key].append(
                        (target_type, target_node_name, priority, fn, input_port_id)
                    )

        # Create shared fn insertion commands according to external hooks collected from
        # pre/post layer, pre/post hooks and op with weights target points.
        for module_type_name, storage in context_hooks.items():
            for storage_key, call_hook_list_info in storage.items():
                compression_module = getattr(self, module_type_name)[storage_key]
                target_points = []
                for target_type, target_node_name, priority, fn, input_port_id in call_hook_list_info:
                    target_points.append(PTTargetPoint(target_type, target_node_name, input_port_id=input_port_id))

                if module_type_name == EXTERNAL_QUANTIZERS_STORAGE_NAME:
                    module_type = ExtraCompressionModuleType.EXTERNAL_QUANTIZER
                elif module_type_name == EXTERNAL_OP_STORAGE_NAME:
                    module_type = ExtraCompressionModuleType.EXTERNAL_OP
                else:
                    raise RuntimeError(f"Module type {module_type_name} is not supported")

                command = PTSharedFnInsertionCommand(
                    target_points=target_points,
                    fn=compression_module,
                    op_unique_name=storage_key,
                    compression_module_type=module_type,
                    priority=priority,
                )
                transformation_layout.register(command)

        return transformation_layout

    def get_node_to_op_address_mapping(self) -> Dict[NNCFNodeName, OperationAddress]:
        """
        Returns map of NNCFGraph node names vs DynamicGraph operation addresses.

        :return: NNCFGraph node names vs DynamicGraph operation addresses map.
        """
        graph_pair = self._compressed_graphs_pair
        if graph_pair is None:
            graph_pair = self._original_graphs_pair

        retval = {}
        for node in graph_pair.dynamic_graph.get_all_nodes():
            node_id = node.node_id
            op_address = node.op_exec_context.op_address
            nncf_node = graph_pair.nncf_graph.get_node_by_id(node_id)
            retval[nncf_node.node_name] = op_address
        return retval

    def get_op_address_to_op_name_map(self) -> Dict[OperationAddress, NNCFNodeName]:
        """
        Returns map of DynamicGraph operation addresses vs NNCFGraph node names.

        :return: DynamicGraph operation addresses vs NNCFGraph node names.
        """
        retval = defaultdict(list)
        for nncf_node_name, op_address in self.get_node_to_op_address_mapping().items():
            retval[op_address].append(nncf_node_name)
        return retval

    def set_compression_controller(self, ctrl: CompressionAlgorithmController):
        self.compression_controller = ctrl

    def strip(self, do_copy: bool = True) -> "NNCFNetwork":
        """
        Returns the model object with as much custom NNCF additions as possible removed
        while still preserving the functioning of the model object as a compressed model.
        :param do_copy: If True (default), will return a copy of the currently associated model object. If False,
          will return the currently associated model object "stripped" in-place.
        :return: The stripped model.
        """
        if self.compression_controller is None:
            # PTQ algorithm does not set compressed controller
            from nncf.torch.quantization.strip import strip_quantized_model

            model = deepcopy(self._model_ref) if do_copy else self._model_ref
            return strip_quantized_model(model)
        return self.compression_controller.strip(do_copy)

    def get_reused_parameters(self):
        """
        Return a list of parameter names which are used as an input in several operations of the model.

        :return: A list of parameter names.
        """
        ret = []
        graph = self._original_graphs_pair.nncf_graph
        for node in graph.get_nodes_by_metatypes(CONST_NOOP_METATYPES):
            if node.is_shared():
                ret.append(node.layer_attributes.name)
        return ret


class NNCFNetworkMeta(type):
    """
    Metaclass for the NNCFNetwork mixin. Has magic methods defined so that the original model object could be
    extended with NNCF-related functionality via a conventional `nncf_network = NNCFNetwork(original_model, ...)`
    syntax and at the same time retain its original class so that downstream class-based checks for the original
    model type don't fail.
    """

    def __call__(
        cls,
        original_model: torch.nn.Module,
        input_info: ModelInputInfo = None,
        dummy_forward_fn: Callable = None,
        wrap_inputs_fn: WrapInputsFnType = None,
        scopes_without_shape_matching: List[str] = None,
        ignored_scopes: List[str] = None,
        target_scopes: List[str] = None,
        wrap_outputs_fn: WrapOutputsFnType = None,
        replace_modules: bool = True,
        trace_parameters: bool = False,
    ) -> "NNCFNetwork":
        """
        This function plays the role of a "constructor" call in the `nncf_network = NNCFNetwork(original_model, ...)`
        syntax. *_scopes arguments are to be passed as string representation of either
        `nncf.common.graph.graph.NNCFNodeName` or `nncf.torch.dynamic_graph.scope.Scope` objects.
        :param original_model: The original model object to be extended with NNCF functionality.
        :param input_infos: A list of descriptors of each tensor input to the model. Will be used to properly generate
            dummy inputs during internal forward calls of the original model for purposes of control flow graph
            building.
        :param dummy_forward_fn: A function to be called instead of the model's original forward function during
            control flow graph building.
        :param wrap_inputs_fn: A user-defined function that will be called with the model's forward arguments
            at each call of the NNCFNetwork object and within which the
            `nncf.torch.dynamic_graph.io_handling.nncf_model_input` function is expected to be called upon each tensor
            among the arguments that is to be treated as an input tensor to the model, thus overriding `input_infos`.
        :param scopes_without_shape_matching: A list of scopes in the model in which the activation tensor shapes will
            not be considered for purposes of scope matching - this helps handle RNN-like cases.
        :param ignored_scopes: A list of scopes in the model for which NNCF handling should not be applied.
            Functions as a "denylist". If left unspecified, nothing will be ignored.
        :param target_scopes: A list of scopes in the model for which NNCF handling should be applied. Functions as
            an "allowlist". If left unspecified, everything will be targeted.
        :param wrap_outputs_fn: Same as `wrap_inputs_fn`, but for marking model outputs with
            `nncf.torch.dynamic_graph.io_handling.nncf_model_output` calls.
        :param replace_modules: Whether to replace model modules with NNCF modules. Default is True.
        :param trace_parameters: Whether to trace model parameters. Default is False.
        :return: The same object as passed in `original_model`, but with internal modules extended/replaced for
            purposes of further NNCF compression, and its class dynamically extended with the `NNCFNetwork`
            as a base class. The object will pass both isinstance(retval, original_model.__class__) and
            isinstance(retval, NNCFNetwork) checks.
        """
        original_class = original_model.__class__
        original_model._nncf = NNCFNetworkInterface(
            original_model,
            input_info,
            dummy_forward_fn,
            wrap_inputs_fn,
            scopes_without_shape_matching,
            ignored_scopes,
            target_scopes,
            wrap_outputs_fn,
            replace_modules,
            trace_parameters,
        )
        # The new class will also have an adjusted metaclass to avoid a "metaclass conflict" upon
        # class creation
        original_metaclass = type(original_model.__class__)
        class_creation_kwds = {}
        if original_metaclass is not type:
            new_metaclass = types.new_class(
                name=original_metaclass.__name__, bases=(NNCFNetworkMeta, original_metaclass)
            )
            class_creation_kwds["metaclass"] = new_metaclass
        new_class = types.new_class(
            name=original_model.__class__.__name__,
            bases=(NNCFNetwork, original_model.__class__),
            kwds=class_creation_kwds,
        )
        # Make the signature of the forward on the resulting object same as for
        # the original forward.
        new_class.forward = _get_nncf_forward_function_with_signature(inspect.signature(original_class.forward))

        # In case of overriding forward by code like `model.forward = wrapper(model.forward)`
        original_instance_forward = original_model.__dict__.get("forward")
        if original_instance_forward is not None:
            bound_new_instance_forward = _get_nncf_forward_function_with_signature(
                inspect.signature(original_instance_forward), bind_self_to=original_model
            )
            original_model.__dict__["forward"] = bound_new_instance_forward

        # Make resulting class keep __module__ attributes of the original class,
        # otherwise these will point to NNCF
        new_class.__module__ = original_class.__module__
        original_model.__class__ = new_class

        if isinstance(original_model, torch.nn.Sequential):
            # If the top-level module is Sequential, then the addition of the ._nncf module will result in
            # the NNCFInterface module being iterated over during the torch.nn.Sequential call, with an attempt to call
            # its forward method, which it effectively doesn't have. Employing a special iterator allows to hide the
            # NNCFInterface object during iteration.
            def nncf_safe_iter(self: torch.nn.Sequential):
                return NNCFSkippingIter(iter(self._modules.values()))

            original_model.__class__.__iter__ = nncf_safe_iter
        return original_model

    def __hash__(cls):
        """
        Makes the dynamically created class object for the processed original model object return the same value when
        hashed as the original class. This allows to gracefully handle the situation when the downstream training
        pipeline checks that the model's class is registered in some registry and determines a training approach
        based on that.
        """
        # expected from a compressed model object to have a NNCFNetwork as 0-th base
        # and original class as 1-st
        if len(cls.__bases__) == 2:
            original_class = cls.__bases__[1]
            return hash(original_class)
        return id(NNCFNetwork)  # conforms to a default hashing behavior in Python for cls objects

    def __eq__(cls, other):
        """
        Makes the dynamically created class object for the processed original model object compare equal with the
        original class object. This allows to gracefully handle the situation when the downstream training
        pipeline checks that the model's class is registered in some registry and determines a training approach
        based on that.
        """
        if len(cls.__bases__) == 2:
            original_class = cls.__bases__[1]
            return original_class == other
        return other is NNCFNetwork


def _get_nncf_forward_function_with_signature(
    signature: inspect.Signature, bind_self_to: torch.nn.Module = None
) -> Callable:
    """
    Creates a function that executes code from NNCFNetwork.forward, but with a final signature equal to the provided
     one.
    :param signature: Signature of function that will used for forward function.
    :param bind_self_to: If provided, will bind the `self` argument of the returned function to the provided model
      object. This should be the model object that we are currently constructing the NNCFNetwork with.
    :return: New copy of function NNCFNetwork.forward with specified signature.
    """
    fn = NNCFNetwork.forward
    new_forward = types.FunctionType(fn.__code__, fn.__globals__, fn.__name__, fn.__defaults__, fn.__closure__)
    new_forward.__dict__.update(fn.__dict__)
    if bind_self_to is not None:
        new_forward = functools.partial(new_forward, bind_self_to)
    new_forward.__signature__ = signature
    if is_debug():
        new_forward = debuggable_forward(new_forward)
    return new_forward


class NNCFNetwork(torch.nn.Module, metaclass=NNCFNetworkMeta):
    """
    A mixin-like class to dynamically extend the original model object's class with.
    """

    TRACE_PARAMETERS_KEY = "trace_parameters"

    def __init__(self, *args, **kwargs):
        """
        In normal situations, the __init__ of the NNCFNetwork will never be called. The constructor-like syntax is
        achieved by a __call__ method defined in the metaclass `NNCFNetworkMeta`.
        """
        super().__init__()
        raise nncf.InternalError("Direct instantiation of NNCFNetwork objects using __init__ is prohibited.")

    def __call__(self, *args, **kwargs):
        """
        Ensures that functor-like calls of the processed model object will directly trigger the NNCF-specific
        forward call.
        """
        return ORIGINAL_CALL(self, *args, **kwargs)

    def forward(self, *args, **kwargs):
        """
        Wraps the original forward call, doing additional actions before and after the call to facilitate model
        graph tracing and calling compression-related hooks.
        """

        with self.nncf._compressed_context as ctx:
            ctx.base_module_thread_local_replica = self

            # add tracing capabilities to model parameters
            if self.nncf.trace_parameters:
                wrap_parameters(self)

            args, kwargs = replicate_same_tensors((args, kwargs))
            if not self.nncf._in_user_dummy_forward:
                # If a user supplies own dummy forward, he is responsible for
                # correctly wrapping inputs inside it as well.
                args, kwargs = strip_traced_tensors(args, kwargs)
                args, kwargs = self.nncf._wrap_inputs_fn(args, kwargs)

            # For purposes of scope tracking, need the original forward call to occur as if it were
            # a module call of the corresponding object.
            if self.nncf._original_instance_forward is not None:

                def _unbound_like_original_instance_forward(_self, *args, **kwargs):
                    return self.nncf._original_instance_forward(*args, **kwargs)

                retval = wrap_module_call(_unbound_like_original_instance_forward)(self, *args, **kwargs)

            elif self.nncf._bound_original_forward is None:
                retval = wrap_module_call(self.nncf._original_unbound_forward)(self, *args, **kwargs)
            else:

                def _unbound_like_original_forward(_self, *args, **kwargs):
                    return self.nncf._bound_original_forward(*args, **kwargs)

                retval = wrap_module_call(_unbound_like_original_forward)(self, *args, **kwargs)

            retval = replicate_same_tensors(retval)
            if not self.nncf._in_user_dummy_forward:
                retval = self.nncf._wrap_outputs_fn(retval)

        if self.nncf._kd_loss_handler is not None and self.training:
            self.nncf._kd_loss_handler(retval, *args, **kwargs)
        return retval

    @property
    def nncf(self) -> NNCFNetworkInterface:
        """
        Accessor for all NNCF-specific methods and attributes of the compressed model object.
        """
        # self._nncf is being set in the creation function defined in the NNCFNetworkMeta metaclass
        return self._nncf

    def __setattr__(self, key, value):
        # If setting `forward`, set it on the original model.
        if key == "forward":
            nncf_logger.warning(
                "You are setting `forward` on an NNCF-processed model object.\n"
                "NNCF relies on custom-wrapping the `forward` call in order to function properly.\n"
                "Arbitrary adjustments to the forward function on an NNCFNetwork object have undefined behavior.\n"
                "If you need to replace the underlying forward function of the original model so that "
                "NNCF should be using that instead of the original forward function that NNCF saved "
                "during the compressed model creation, you can do this by calling:\n"
                "model.nncf.set_original_unbound_forward(fn)\n"
                "if `fn` has an unbound 0-th `self` argument, or\n"
                "with model.nncf.temporary_bound_original_forward(fn): ...\n"
                "if `fn` already had 0-th `self` argument bound or never had it in the first place."
            )
        super().__setattr__(key, value)


class NNCFSkippingIter:
    """
    An iterator over the regular torch.nn.Module iterator that will skip NNCFInterface objects if they come up.
    """

    def __init__(self, iter_to_wrap: Iterator[Module]):
        self._iter_to_wrap = iter_to_wrap

    def __next__(self):
        item = next(self._iter_to_wrap)
        if isinstance(item, NNCFNetworkInterface):
            item = next(self._iter_to_wrap)
        return item


class LoadStateListener:
    """
    Resets the initialization flags (`initialized`) for all quantization modules on `load_state_dict` call.
    These flags are used to update not loaded params (from checkpoint or model's state)
    on initialization stage of algorithm.
    Flags reset is required on each call of `load_state_dict`, because internal method (`build_graph`)
    restores model state by calling this method.
    """

    def __init__(self, model: "NNCFNetwork", all_quantizations: Dict[str, torch.nn.Module]):
        self.hook = model._register_load_state_dict_pre_hook(
            functools.partial(self.hook_fn, quantize_modules=list(all_quantizations.values()))
        )

    def hook_fn(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs,
        quantize_modules: List[torch.nn.Module],
    ):
        for module in quantize_modules:
            module.initialized = False

    def close(self):
        self.hook.remove()


def compression_module_type_to_attr_name(compression_module_type: ExtraCompressionModuleType):
    """
    Required for backward compatibility with checkpoints that store function and activation
    quantizers directly under corresponding attributes of NNCFNetwork.
    """
    if compression_module_type == ExtraCompressionModuleType.EXTERNAL_QUANTIZER:
        return EXTERNAL_QUANTIZERS_STORAGE_NAME
    if compression_module_type == ExtraCompressionModuleType.EXTERNAL_OP:
        return EXTERNAL_OP_STORAGE_NAME
    raise nncf.ValidationError("Unknown extra module type")
